(* Nova-Simulator
   ==============

    J. Rathlev, Schwentinental, (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Apr. 2007
   Vers. 2 - August 2017 : compiled under Delphi 10 Seattle
   last modified:  May 2020
   *)


unit NovaMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ComCtrls,
  Vcl.Samples.Spin, Vcl.Menus, Vcl.Grids, System.Contnrs, Vcl.ExtCtrls, Vcl.Buttons,
  System.IniFiles, WinApiUtils, NumberEd, Indicators, FpNova3D, Switches,
  Vcl.WinXCtrls, NumberUtils;

const
  ProgName = 'Nova-Simulator';
  Vers = ' - Vers. 2.4';
  CopRgt = ' 2007-2020 Dr. J. Rathlev';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  maxMemSize = 32768;
  MemStackSize = 64;
  StartAdr = $2;
  TrailerLength = 100;   // trailing zeros on PTR
  defSwitches = '000012,177777'; // presetted switch settings

  iniTop = 'Top';
  iniLeft = 'Left';
  iniHeight = 'Height';
  iniWidth = 'Width';


  DmpExt = 'dmp';
  LstExt = 'lst';

  nPrgLoad = 32;
  PrgLoad    : array [0..nPrgLoad-1] of word =
                         (26047,24895,10262,36608,43008,04108,04120,04122,
                          43780,00005,12302,20735,24639,33346,00255,02072,
                          33333,00015,02071,19478,04160,00018,00063,44304,
                          26495,00024,24895,36595,00024,43712,00768,00000);

  RTCIntervals : array [0..3] of integer = (20,100,10,1);

type
  TAkkus = array [0..3] of word;
  TBreakPoint = (bpNormal,bpStepOver);
  TBreakPoints = set of TBreakPoint;

  TMemory = record
    Value : word;
    BreakPoint : TBreakPoints;
    end;

const
  FirstDevice = 2;
  LastDevice = 62;

type
  TByteOut = procedure(Sender: TObject; AByte : byte) of object;

  TTraceData = class(TObject)
    PC,Data,Carry : word;
    ACs : TAkkus;
    constructor Create (APC,AData,ACry : word; Akkus : TAkkus);
    end;

  TWatchMem = class(TObject)
    Addr,Data : word;
    Alert,Changed : boolean;
    constructor Create (AAddr,AData : word);
    end;

  TInputStream = class (TFileStream)
    FName : string;
    TrailerCount : integer;
    constructor Create (const AFilename : string);
    end;

  TDevice = class(TObject)
  private
    FBusy,FDone,FMask,FIrq,
    FCompleted             : boolean;
    FAdr,FMaskBit          : word;
    FTickCount,FTick       : integer;
    FOnInterrupt           : TNotifyEvent;
  public
    constructor Create (Adr,Bit,Ticks : word);
    destructor Destroy; override;
    procedure ProcessTick; virtual;
    procedure Reset; virtual;
    procedure Complete; virtual;
    function DIA : word; virtual;
    function DIB : word; virtual;
    function DIC : word; virtual;
    procedure DOA (Value : word); virtual;
    procedure DOB (Value : word); virtual;
    procedure DOC (Value : word); virtual;
    procedure Clear; virtual;
    procedure Start; virtual;
    procedure Pulse; virtual;
    procedure MaskOut (Mask : word);
    property BusyFlag : boolean read FBusy;
    property DoneFlag : boolean read FDone;
    property Interrupt : boolean read FIrq;
    property OnInterrupt : TNotifyEvent read FOnInterrupt write FOnInterrupt;
    end;

  TRtcDevice = class(TDevice)
  private
    FTimer : TTimer;
  public
    constructor Create (Adr,Bit : word; ATimer : TTimer);
    procedure DOA (Value : word); override;
    procedure Clear; override;
    procedure Start; override;
    procedure TimeElapsed;
    end;

  TOutDevice = class(TDevice)
  private
    OByte: byte;
    FOnByteOut : TByteOut;
  public
    constructor Create (Adr,Bit,Ticks : word);
    procedure Reset; override;
    procedure DOA (Value : word); override;
    procedure Start; override;
    property OnByteOut : TByteOut read FOnByteOut write FOnByteOut;
    end;

  TPrintDevice = class(TOutDevice)
  public
    function DIA : word; override;
    procedure DOA (Value : word); override;
    end;

  TInDevice = class(TDevice)
  private
    IByte : byte;
    FOnByteRequest : TNotifyEvent;
    procedure SetInputByte(Value : byte);
  public
    constructor Create (Adr,Bit,Ticks : word);
    procedure Reset; override;
    function DIA : word; override;
    procedure Start; override;
    property InputByte: byte read IByte write SetInputByte;
    property OnByteRequest : TNotifyEvent read FOnByteRequest write FOnByteRequest;
    end;

  TTeletypeInput = class(TInDevice)
  private
    FRdOn : boolean;
    procedure SetReaderOn (Value : boolean);
  public
    constructor Create (Adr,Bit,Ticks : word);
    procedure ProcessTick; override;
    procedure Start; override;
    procedure Reset; override;
    procedure Complete; override;
    property ReaderOn : boolean read FRdOn write SetReaderOn;
    end;

  TMainForm = class(TForm)
    gbRegister: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    edPC: TNumberEdit;
    edCarry: TNumberEdit;
    edAC0: TNumberEdit;
    edAC1: TNumberEdit;
    edAC2: TNumberEdit;
    edAC3: TNumberEdit;
    rbBin: TRadioButton;
    rbOctal: TRadioButton;
    rbHex: TRadioButton;
    rbDecimal: TRadioButton;
    gbMem: TGroupBox;
    GroupBox3: TGroupBox;
    cbPC: TCheckBox;
    cbAC0: TCheckBox;
    cbAC1: TCheckBox;
    cbAC2: TCheckBox;
    cbAC3: TCheckBox;
    MainMenu: TMainMenu;
    itmDump: TMenuItem;
    itmPrefs: TMenuItem;
    itInfo: TMenuItem;
    itEnde: TMenuItem;
    itLoadDump: TMenuItem;
    itSaveDump: TMenuItem;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    itmWorkDir: TMenuItem;
    sgMemory: TStringGrid;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    GroupBox1: TGroupBox;
    btRun: TSpeedButton;
    Panel4: TPanel;
    Label7: TLabel;
    edCmd: TNumberEdit;
    cbCmd: TCheckBox;
    aiSkip: TArrowIndicator;
    Label8: TLabel;
    btStep: TSpeedButton;
    btReset: TSpeedButton;
    edAdr: TNumberEdit;
    Label9: TLabel;
    itmMemSize: TMenuItem;
    edCommand: TEdit;
    mmTTO: TMemo;
    btStop: TSpeedButton;
    Label10: TLabel;
    Label11: TLabel;
    siIntEn: TSpotIndicator;
    siIrq: TSpotIndicator;
    itmInOut: TMenuItem;
    itmPTP: TMenuItem;
    itmPTR: TMenuItem;
    N1: TMenuItem;
    itmDevices: TMenuItem;
    itmPtrStart: TMenuItem;
    itmPtrStop: TMenuItem;
    btPrgLoad: TSpeedButton;
    N3: TMenuItem;
    itWritePrg: TMenuItem;
    GroupBox4: TGroupBox;
    cbSwitches: TCheckBox;
    edSwitches: TNumberEdit;
    tmRtc: TTimer;
    pmMemory: TPopupMenu;
    itpSetPC: TMenuItem;
    itpEdit: TMenuItem;
    N4: TMenuItem;
    itpCancel: TMenuItem;
    itmPtrReset: TMenuItem;
    itmMemReset: TMenuItem;
    btStepOver: TSpeedButton;
    itmFrontpanel: TMenuItem;
    N5: TMenuItem;
    itmLang: TMenuItem;
    itmGerman: TMenuItem;
    itmEnglish: TMenuItem;
    itpToggleBp: TMenuItem;
    itmActions: TMenuItem;
    itmReset: TMenuItem;
    itmRun: TMenuItem;
    itmStep: TMenuItem;
    itmStepover: TMenuItem;
    itmStop: TMenuItem;
    btPtrReset: TBitBtn;
    btPtrOpen: TBitBtn;
    btPtrClose: TBitBtn;
    edPtp: TEdit;
    btPtpOpen: TBitBtn;
    btPtpClose: TBitBtn;
    GroupBox2: TGroupBox;
    edMemStep: TNumberEdit;
    btmemUp: TBitBtn;
    btmemDown: TBitBtn;
    Label13: TLabel;
    Label14: TLabel;
    edPtrTotal: TNumberEdit;
    edPtrPos: TNumberEdit;
    Label16: TLabel;
    edPtpPos: TNumberEdit;
    btSetPC: TSpeedButton;
    itpSeek: TMenuItem;
    Panel6: TPanel;
    btLast: TSpeedButton;
    btNext: TSpeedButton;
    Label15: TLabel;
    siRun: TSpotIndicator;
    btSwitchesToPC: TSpeedButton;
    Label17: TLabel;
    edPtrByte: TNumberEdit;
    Label18: TLabel;
    edPtpByte: TNumberEdit;
    cbUpcase: TCheckBox;
    btShowSteps: TSpeedButton;
    itmView: TMenuItem;
    itmShowTrace: TMenuItem;
    itmShowWatch: TMenuItem;
    itpWatch: TMenuItem;
    N2: TMenuItem;
    itpGoto: TMenuItem;
    N6: TMenuItem;
    itmReload: TMenuItem;
    N7: TMenuItem;
    itmWatchlocation: TMenuItem;
    itmBreakpoints: TMenuItem;
    N8: TMenuItem;
    btPC: TSpeedButton;
    N9: TMenuItem;
    itmTextfiles: TMenuItem;
    Panel7: TPanel;
    btQuit: TBitBtn;
    btInfo: TBitBtn;
    btFrontPanel: TBitBtn;
    btTrace: TBitBtn;
    btWatch: TBitBtn;
    btBreakpoints: TBitBtn;
    rbPtrTextFile: TRadioButton;
    rbPtrBinFile: TRadioButton;
    edPtr: TComboBox;
    tmView: TTimer;
    aiRun: TActivityIndicator;
    pcDevices: TPageControl;
    tsTTR: TTabSheet;
    tsPTR: TTabSheet;
    tsPTP: TTabSheet;
    pnPTP: TPanel;
    pnPtr: TPanel;
    tsTTP: TTabSheet;
    tsLPT: TTabSheet;
    pnTTP: TPanel;
    Label19: TLabel;
    Label20: TLabel;
    btTtpOpen: TBitBtn;
    edTtpClose: TBitBtn;
    edTtp: TEdit;
    edTtpByte: TNumberEdit;
    edTtpPos: TNumberEdit;
    pnLPT: TPanel;
    btLptOpen: TBitBtn;
    btLptClose: TBitBtn;
    edLpt: TEdit;
    pnTTR: TPanel;
    Label23: TLabel;
    Label24: TLabel;
    Label25: TLabel;
    edTtrClose: TBitBtn;
    btTtrOpen: TBitBtn;
    edTtrReset: TBitBtn;
    edTtr: TComboBox;
    edTtrByte: TNumberEdit;
    edTtrPos: TNumberEdit;
    edTtrTotal: TNumberEdit;
    rbTtrBinFile: TRadioButton;
    rbTtrTextFile: TRadioButton;
    gbTTY: TGroupBox;
    tswTTReader: TToggleSwitch;
    Label26: TLabel;
    tswTTPunch: TToggleSwitch;
    Label27: TLabel;
    cbPreSwitches: TComboBox;
    itmSwitches: TMenuItem;
    mmLpt: TMemo;
    switchon1: TMenuItem;
    switchoff1: TMenuItem;
    btAddWatch: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure rbBinClick(Sender: TObject);
    procedure rbOctalClick(Sender: TObject);
    procedure rbDecimalClick(Sender: TObject);
    procedure rbHexClick(Sender: TObject);
    procedure cbPCClick(Sender: TObject);
    procedure cbAC0Click(Sender: TObject);
    procedure cbAC1Click(Sender: TObject);
    procedure cbAC2Click(Sender: TObject);
    procedure cbAC3Click(Sender: TObject);
    procedure itLoadDumpClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure itInfoClick(Sender: TObject);
    procedure itSaveDumpClick(Sender: TObject);
    procedure itEndeClick(Sender: TObject);
    procedure cbCmdClick(Sender: TObject);
    procedure btResetClick(Sender: TObject);
    procedure btStepClick(Sender: TObject);
    procedure itmWorkDirClick(Sender: TObject);
    procedure itmMemSizeClick(Sender: TObject);
    procedure edAC3Click(Sender: TObject);
    procedure edAC2Click(Sender: TObject);
    procedure edAC0Click(Sender: TObject);
    procedure edAC1Click(Sender: TObject);
    procedure edCmdClick(Sender: TObject);
    procedure edPCClick(Sender: TObject);
    procedure sgMemoryDblClick(Sender: TObject);
    procedure sgMemoryDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure btRunClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure btStopClick(Sender: TObject);
    procedure itmPtrOpenClick(Sender: TObject);
    procedure itmPtrCloseClick(Sender: TObject);
    procedure cbSwitchesClick(Sender: TObject);
    procedure itWritePrgClick(Sender: TObject);
    procedure edSwitchesClick(Sender: TObject);
    procedure sgMemoryKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btPrgLoadClick(Sender: TObject);
    procedure edCarryClick(Sender: TObject);
    procedure tmRtcTimer(Sender: TObject);
    procedure itpSetPCClick(Sender: TObject);
    procedure itmPtrResetClick(Sender: TObject);
    procedure itmMemResetClick(Sender: TObject);
    procedure sgMemoryTopLeftChanged(Sender: TObject);
    procedure itmFrontpanelClick(Sender: TObject);
    procedure itmGermanClick(Sender: TObject);
    procedure itmEnglishClick(Sender: TObject);
    procedure itpToggleBpClick(Sender: TObject);
    procedure btStepOverClick(Sender: TObject);
    procedure btPtpCloseClick(Sender: TObject);
    procedure btmemUpClick(Sender: TObject);
    procedure btmemDownClick(Sender: TObject);
    procedure btTraceClick(Sender: TObject);
    procedure edAC0KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure edAC1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure edAC2KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure edAC3KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure edCarryKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure itpSeekClick(Sender: TObject);
    procedure sgMemoryMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btLastClick(Sender: TObject);
    procedure btNextClick(Sender: TObject);
    procedure btSwitchesToPCClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure edAdrMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure gbRegisterMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure mmTTOKeyPress(Sender: TObject; var Key: Char);
    procedure itpWatchClick(Sender: TObject);
    procedure btWatchClick(Sender: TObject);
    procedure itpGotoClick(Sender: TObject);
    procedure itmReloadClick(Sender: TObject);
    procedure itmBreakpointsClick(Sender: TObject);
    procedure btPCClick(Sender: TObject);
    procedure itmTextfilesClick(Sender: TObject);
    procedure btShowStepsClick(Sender: TObject);
    procedure edPtrCloseUp(Sender: TObject);
    procedure edPtrDrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
      State: TOwnerDrawState);
    procedure edPtrDropDown(Sender: TObject);
    procedure tmViewTimer(Sender: TObject);
    procedure cbPreSwitchesDropDown(Sender: TObject);
    procedure cbPreSwitchesCloseUp(Sender: TObject);
    procedure itmSwitchesClick(Sender: TObject);
    procedure tswTTReaderClick(Sender: TObject);
    procedure tswTTPunchClick(Sender: TObject);
    procedure btLptOpenClick(Sender: TObject);
    procedure btLptCloseClick(Sender: TObject);
    procedure edTtrResetClick(Sender: TObject);
    procedure btTtrOpenClick(Sender: TObject);
    procedure edTtrCloseClick(Sender: TObject);
    procedure edTtrCloseUp(Sender: TObject);
    procedure edTtrDropDown(Sender: TObject);
    procedure edTtrDrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
      State: TOwnerDrawState);
    procedure edTtpCloseClick(Sender: TObject);
    procedure btPtpOpenClick(Sender: TObject);
    procedure btTtpOpenClick(Sender: TObject);
    procedure btAddWatchClick(Sender: TObject);
  private
    { Private-Deklarationen }
    IniName,
    DumpName,
    ListName,
    LastPtr,
    LastTTR,
    TextFiles,
    LptLine   : string;
    Memory    : array [0..maxMemSize-1] of TMemory;
    MemStack  : array [0..MemStackSize-1] of word;
    MemStackPtr,
    MemStackCnt  : integer;
    FSwitches,
    MemSize,
    FCarry,
    PC        : word;
    Akkus     : TAkkus;
    PreEnable,
    IntEnable,
    IrqFlag,
    FRun,
    ContentsCol,
    Loaded    : boolean;
    NumFormat : TNumMode;
    IoDevices : array [FirstDevice..LastDevice] of TDevice;
    PtrFile,
    TtrFile   : TInputStream;
    PtpFile,
    TtpFile,
    LptFile   : TFileStream;
    HintWin   : THintWindow;
    function CheckValue  (const AText : string) : boolean;
    procedure ShowRegister;
    procedure UpdateView;
    procedure UpdateMemView;
    procedure GotoMem (Adr : integer);
    procedure SaveMemPos (Adr : integer);
    function LoadDump(const FileName : string; ClearBp : boolean) : boolean;
    procedure SaveDump(FileName : string);
    procedure ClearRegs;
    procedure ShowCmd;
    function IntAck : word;
    procedure MaskOut (Value : word);
    procedure IoReset;
    procedure Reset;
    function IndirectAdr (Addr : word; Execute : boolean) : word;
    function ResolveAdr (Command : word; Execute : boolean) : word;
    function AnalyseCmd (cmd : word; Execute : boolean) : boolean;
    procedure ProcessInterrupts;
    procedure RunProg;
    procedure ProgramLoad;
    function NoMoreBytes (fs : TStream) : boolean;
    function OpenTtr (const Filename : string) : boolean;
    function SelectTtrFile : boolean;
    procedure ReadFromTtr (Sender: TObject);
    procedure WriteToTTO (Sender: TObject; AByte : byte);
    function OpenTtp (const Filename : string) : boolean;
    function SelectTtpFile : boolean;
    function OpenPtp (const Filename : string) : boolean;
    function SelectPtpFile : boolean;
    procedure WriteToPTP (Sender: TObject; AByte : byte);
    function OpenLpt (const Filename : string) : boolean;
    function SelectLptFile : boolean;
    procedure WriteToLPT (Sender: TObject; AByte : byte);
    function OpenPtr (const Filename : string) : boolean;
    function SelectPtrFile : boolean;
    procedure ReadFromPtr (Sender: TObject);
    procedure SetInterrupt (Sender: TObject);
    function OpenInputFile (ATitle,AFilter,AExt : string; var AFilename : string) : boolean;
    function OpenOutputFile (ATitle,AFilter,AExt : string; var AFilename : string) : boolean;
    procedure DataSwitchChanged (Value : word);
    procedure CommandSwitchChanged (Sender : TObject; Command : TNovaCommands);
    procedure AddToTrace (TrData : TTraceData);
    procedure AddToWatch (AAddr,AData : word);
    function UpdateWatch (WatchChanges : boolean = false) : boolean;
    procedure AddBreakpoint (AAddr : word);
    procedure RemBreakpoint (AAddr : word);
    procedure DelBreakpoint (Addr : word);
    function SetTextFileFilter (const ExtList : string) : string;
    function HasTextExt (const Filename : string) : boolean;
  public
    { Public-Deklarationen }
    ProgVersName,
    ProgVersDate,ProgVers,
    AppPath,WorkPath      : string;
    Trace     : TObjectList;
    BreakPoints,
    Watch     : TStringList;
    procedure ClearWatches;
    function DeleteWatch (const sAddr : string) : integer;
    function ReadMemory (Addr : word) : word;
    procedure ShowMemory (Addr : word);
    procedure AddWatch (Addr : word);
    procedure SetBreakpoint (Addr : word);
    function DeleteBreakpoint (const sAddr : string) : integer;
    procedure InvalidateMemView;
    end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses System.Math, System.StrUtils, Winapi.ShlObj, FileUtils, StringUtils,
  NovaUtils, WinShell, ShellDirDlg, InitProg,
  InpValue, InpText, InpCommand, WinUtils, WriteProg,
  GnuGettext, LangUtils, ShowTrace, ShowWatch, ShowBreakpoints,
  SelectFromListDlg;

const
  defTextFiles = 'txt,lst,asm,bas';

{ ------------------------------------------------------------------- }
// Trace data
constructor TTraceData.Create (APC,AData,ACry : word; Akkus : TAkkus);
begin
  inherited Create;
  PC:=APC; Data:=AData; Carry:=ACry; ACs:=Akkus;
  end;

// Watch memory
constructor TWatchMem.Create (AAddr,AData : word);
begin
  inherited Create;
  Addr:=AAddr; Data:=AData; Alert:=false; Changed:=false;
  end;

{ ------------------------------------------------------------------- }
// Input stream (PTR and TTR)
constructor TInputStream.Create (const AFilename : string);
begin
  inherited Create(AFilename,fmOpenRead+fmShareDenyWrite);
  FName:=AFilename; TrailerCount:=0;
  end;

{ ------------------------------------------------------------------- }
// Basic IO functions
constructor TDevice.Create (Adr,Bit,Ticks : word);
begin
  inherited Create;
  FAdr:=Adr; FMaskBit:=1 shl (15-Bit);
  FBusy:=false; FDone:=false;
  FMask:=false; FIrq:=false;
  FCompleted:=false;
  FTickCount:=Ticks; FTick:=0;
  FOnInterrupt:=nil;
  end;

destructor TDevice.Destroy;
begin
  inherited Destroy;
  end;

procedure TDevice.ProcessTick;
begin
  if FTick=0 then begin       // set interrupt
    if FBusy and FCompleted then begin
      FDone:=true; FBusy:=false;
      FCompleted:=false;
      end;
    FIrq:=not FMask and FDone;
    if FIrq and assigned(FOnInterrupt) then FOnInterrupt(self);
    end
  else dec(FTick);
  end;

procedure TDevice.Reset;
begin
  Clear;
  FMask:=false;
  end;

procedure TDevice.Clear;
begin
  FBusy:=false; FDone:=false; FIrq:=false; FCompleted:=false;
  end;

procedure TDevice.Start;
begin
  FBusy:=true; FDone:=false; FIrq:=false;
  FTick:=FTickCount;
  end;

procedure TDevice.Pulse;
begin
  end;

procedure TDevice.MaskOut (Mask : word);
begin
  FMask:=(Mask and FMaskBit)<>0;
  FIrq:=not FMask and FDone;
  if FIrq and assigned(FOnInterrupt) then FOnInterrupt(self);
  end;

procedure TDevice.Complete;
begin
  FCompleted:=true;
  end;

function TDevice.DIA : word;
begin
  Result:=0;
  end;

function TDevice.DIB : word;
begin
  Result:=0;
  end;

function TDevice.DIC : word;
begin
  Result:=0;
  end;

procedure TDevice.DOA (Value : word);
begin
  end;

procedure TDevice.DOB (Value : word);
begin
  end;

procedure TDevice.DOC (Value : word);
begin
  end;

{ ------------------------------------------------------------------- }
// Real Time Clock
constructor TRtcDevice.Create (Adr,Bit : word; ATimer : TTimer);
begin
  inherited Create (Adr,Bit,0);
  FTimer:=ATimer;
  end;

procedure TRtcDevice.DOA (Value : word);
begin
  if assigned(FTimer) then FTimer.Interval:=RTCIntervals[Value and 3];
  end;

procedure TRtcDevice.Clear;
begin
  if assigned(FTimer) then FTimer.Enabled:=false;
  inherited Clear;
  end;

procedure TRtcDevice.Start;
begin
  if assigned(FTimer) then FTimer.Enabled:=true;
  inherited Start;
  end;

procedure TRtcDevice.TimeElapsed;
begin
  if FBusy then begin
    FDone:=true; FBusy:=false;
    FCompleted:=true;
    end;
  end;

{ ------------------------------------------------------------------- }
// Output device for byte
constructor TOutDevice.Create (Adr,Bit,Ticks : word);
begin
  inherited Create (Adr,Bit,Ticks);
  OByte:=0;
  FOnByteOut:=nil;
  end;

procedure TOutDevice.Reset;
begin
  OByte:=0;
  end;

procedure TOutDevice.DOA (Value : word);
begin
  OByte:=Value and $FF;
  end;

procedure TOutDevice.Start;
begin
  inherited Start;
  if Assigned(FOnByteOut) then FOnByteOut(self,OByte);
  Complete;
  end;

{ ------------------------------------------------------------------- }
procedure TPrintDevice.DOA (Value : word);
begin
  OByte:=Value and $7F;
  if (OByte<32) and (Obyte<>$D) and (Obyte<>$A) and (Obyte<>$C) then Obyte:=32; // set to space
  end;

function TPrintDevice.DIA : word;
begin
  Result:=1;   // printer ready
  end;

{ ------------------------------------------------------------------- }
// Input device for byte
constructor TInDevice.Create (Adr,Bit,Ticks : word);
begin
  inherited Create (Adr,Bit,Ticks);
  IByte:=0;
  FOnByteRequest:=nil;
  end;

procedure TInDevice.Reset;
begin
  IByte:=0;
  end;

procedure TInDevice.SetInputByte(Value : byte);
begin
  IByte:=Value;
  Complete;
  end;

function TInDevice.DIA : word;
begin
  Result:=IByte;
  end;

procedure TInDevice.Start;
var
  b : boolean;
begin
  b:=FBusy;
  inherited Start;
  if not b and Assigned(FOnByteRequest) then FOnByteRequest(self);
  end;

{ ------------------------------------------------------------------- }
constructor TTeletypeInput.Create (Adr,Bit,Ticks : word);
begin
  inherited Create (Adr,Bit,Ticks);
  FRdOn:=false;
  end;

procedure TTeletypeInput.SetReaderOn (Value : boolean);
begin
  if FRdOn<>Value then begin
    FRdOn:=Value;
    Start;
    end;
  end;

procedure TTeletypeInput.ProcessTick;
begin
  if FRdOn then inherited ProcessTick;
  end;

procedure TTeletypeInput.Start;
begin
  FBusy:=true; FDone:=false; FIrq:=false;
  if FRdOn then begin
    if Assigned(FOnByteRequest) then FOnByteRequest(self);
    FTick:=FTickCount;
    end;
  end;

procedure TTeletypeInput.Reset;
begin
  inherited Reset;
  FRdOn:=false;
  end;

procedure TTeletypeInput.Complete;
begin
  if FRdOn then inherited Complete
  else begin
    FDone:=true; FBusy:=false;
    FIrq:=not FMask and FDone;
    if FIrq and assigned(FOnInterrupt) then FOnInterrupt(self);
    end;
  end;

{ ------------------------------------------------------------------- }
const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  FpSekt = 'FrontPanel';
  TraceSekt = 'TraceWindow';
  WatchSekt = 'WatchWindow';
  BpSekt = 'Breakpoints';
  PtrSekt  = 'PtrFiles';
  TtrSekt  = 'TtrFiles';

  (* INI-Varibalen *)
  IniStdPath = 'WorkPath';
  IniLastDmp = 'LastDump';
  IniLastLst = 'LastListing';
  IniNumBase = 'NumBase';
  IniMemSz   = 'MemSize';
  IniTtr     = 'TTRName';
  IniTtp     = 'TTPName';
  IniPtr     = 'PTRName';
  IniPtp     = 'PTPName';
  IniLpt     = 'LPTName';
  IniSwitch  = 'Switches';
  IniFrontPan = 'ShowFrontPanel';
  IniTrace    = 'ShowTrace';
  IniWatch    = 'ShowWatch';
  IniUpcase   = 'UpperCase';
  IniVisible  = 'Visible';
  IniCount    = 'Count';
  IniAddr     = 'Address';
  IniTxtFile  = 'TextFiles';
  IniMemStep  = 'MemoryStep';
  IniPreSwtch = 'PreSwitches';

procedure TMainForm.FormCreate(Sender: TObject);
var
  IniFile : TIniFile;
  i,n,na  : integer;
  s       : string;
begin
  TranslateComponent(self);
  InitPaths(AppPath,WorkPath);
  InitVersion(ProgName,Vers,CopRgt,3,3,ProgVersName,ProgVers,ProgVersDate);
  if AnsiSameText(SelectedLanguage,'de') then itmGerman.Click
  else itmEnglish.Click;
  Trace:=TObjectList.Create;
  Watch:=TStringList.Create;
  Watch.Sorted:=true;
  Breakpoints:=TStringList.Create;
  Breakpoints.Sorted:=true;
  IniName:=Erweiter(AppPath,PrgName,IniExt);
// Ini-Datei lesen
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    Top:=ReadInteger(CfgSekt,iniTop,Top);
    Left:=ReadInteger(CfgSekt,iniLeft,Left);
    WorkPath:=ReadString(CfgSekt,IniStdPath,WorkPath);
    DumpName:=ReadString(CfgSekt,IniLastDmp,'');
    ListName:=ReadString(CfgSekt,IniLastLst,'');
    NumFormat:=TNumMode(ReadInteger(CfgSekt,IniNumBase,integer(nmOctal)));
    MemSize:=ReadInteger(CfgSekt,IniMemSz,maxMemSize);
    edMemstep.Value:=ReadInteger(CfgSekt,IniMemStep,256);
    itmFrontpanel.Checked:=ReadBool(CfgSekt,IniFrontPan,true);
    itmShowTrace.Checked:=ReadBool(TraceSekt,IniVisible,false);
    itmShowWatch.Checked:=ReadBool(WatchSekt,IniVisible,false);
    itmBreakpoints.Checked:=ReadBool(BpSekt,IniVisible,false);
    cbUpcase.Checked:=ReadBool(CfgSekt,IniUpcase,true);
    FSwitches:=ReadInteger(CfgSekt,IniSwitch,0);
    cbPreSwitches.Items.CommaText:=ReadString(CfgSekt,IniPreSwtch,defSwitches);
    edLpt.Text:=ReadString(CfgSekt,IniLpt,'');
    edPtp.Text:=ReadString(CfgSekt,IniPtp,'');
    LoadHistory(IniFile,PtrSekt,'',edPtr.Items);
    s:=ReadString(CfgSekt,IniPtr,'');
    with edPtr do begin
      n:=Items.IndexOf(s);
      if n>=0 then ItemIndex:=n
      else if length(s)>9 then AddToHistory(Items,s)
      else ItemIndex:=-1;
      end;
    edTtp.Text:=ReadString(CfgSekt,IniTtp,'');
    LoadHistory(IniFile,TtrSekt,'',edTtr.Items);
    s:=ReadString(CfgSekt,IniTtr,'');
    with edTtr do begin
      n:=Items.IndexOf(s);
      if n>=0 then ItemIndex:=n
      else if length(s)>9 then AddToHistory(Items,s)
      else ItemIndex:=-1;
      end;
    TextFiles:=ReadString(CfgSekt,IniTxtFile,defTextFiles);
    n:=ReadInteger(WatchSekt,IniCount,0);
    for i:=0 to n-1 do begin
      na:=ReadInteger(WatchSekt,IniAddr+ZStrint(i,3),-1);
      if (na>=0) and (na<maxMemSize) then AddToWatch(na,0);
      end;
    n:=ReadInteger(BpSekt,IniCount,0);
    for i:=0 to n-1 do begin
      na:=ReadInteger(BpSekt,IniAddr+ZStrint(i,3),-1);
      if (na>=0) and (na<maxMemSize) then AddBreakpoint(na);
      end;
    Free;
    end;
  case NumFormat of
  nmBin : rbBin.Checked:=true;
  nmOctal : rbOctal.Checked:=true;
  nmHex : rbHex.Checked:=true;
  else rbDecimal.Checked:=true;
    end;
  with sgMemory do begin
    RowCount:=MemSize+1;
    ColWidths[0]:=16; ColWidths[1]:=60; ColWidths[2]:=60;
    ColWidths[3]:=Width-161;
    Canvas.Font:=Font;
    end;
  Loaded:=false;
  HintWin:=THintWindow.Create(self);
  // Ein-Ausgabegerte
  PtrFile:=nil; TtrFile:=nil; PtpFile:=nil; TtpFile:=nil; LptFile:=nil;
  for i:=FirstDevice to LastDevice do IoDevices[i]:=nil;
  IoDevices[ioTTO]:=TOutDevice.Create(ioTTO,mbTTO,tcTTO);
  (IoDevices[ioTTO] as TOutDevice).OnByteOut:=WriteToTTO;
  IoDevices[ioTTI]:=TTeletypeInput.Create(ioTTI,mbTTI,tcTTI);
  (IoDevices[ioTTI] as TTeletypeInput).OnByteRequest:=ReadFromTtr;
  IoDevices[ioPTP]:=TOutDevice.Create(ioPTP,mbPTP,tcPTP);
  (IoDevices[ioPTP] as TOutDevice).OnByteOut:=WriteToPTP;
  IoDevices[ioPTR]:=TInDevice.Create(ioPTR,mbPTR,tcPTR);
  (IoDevices[ioPTR] as TInDevice).OnByteRequest:=ReadFromPtr;
  IoDevices[ioRTC]:=TRtcDevice.Create(ioRTC,mbRTC,tmRtc);
  IoDevices[ioLPT]:=TPrintDevice.Create(ioLPT,mbLPT,tcLPT);
  (IoDevices[ioLPT] as TOutDevice).OnByteOut:=WriteToLPT;
  for i:=FirstDevice to LastDevice do
    if assigned(IoDevices[i]) then IoDevices[i].OnInterrupt:=SetInterrupt;
  gbMem.Caption:=_('Program memory')+' ('+IntToStr(MemSize)+' '+_('words')+' = '+SizeToStr(2*MemSize)+')'; //'Programmspeicher'
  MemStackPtr:=0; MemStackCnt:=0;
  LptLine:='';
  btLast.Enabled:=false;
  btNext.Enabled:=false;
  ContentsCol:=false;
  end;

procedure TMainForm.FormActivate(Sender: TObject);
begin
  with HintWin do begin
    Canvas.Font:=edAdr.Font;
    Brush.Color:=clInfoBk;
    end;
  end;

procedure TMainForm.FormDeactivate(Sender: TObject);
begin
  if Active then HintWin.ReleaseHandle;
  end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
  i : integer;
begin
  for i:=FirstDevice to LastDevice do if assigned(IoDevices[i]) then IoDevices[i].Free;
  ClearWatches; Watch.Free; Breakpoints.Free;
  Trace.Free; HintWin.Free;
  end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  with TraceWin do begin
    LoadFromIni(IniName,TraceSekt);
    Visible:=itmShowTrace.Checked;
    end;
  with WatchWin do begin
    LoadFromIni(IniName,WatchSekt);
    Visible:=itmShowWatch.Checked;
    if Visible then UpdateView(Watch.Count);
    end;
  with BpWin do begin
    LoadFromIni(IniName,BpSekt);
    Visible:=itmBreakpoints.Checked;
    if Visible then UpdateView(Breakpoints.Count);
    end;
  if (length(DumpName)>0) then Loaded:=LoadDump(DumpName,false);
  with NovaFrontpanel do begin
    LoadFromIni(IniName,FpSekt);
    OnChangeDataSwitch:=DataSwitchChanged;
    OnChangeCommandSwitch:=CommandSwitchChanged;
    Visible:=itmFrontPanel.Checked; // Show;
    end;
  if not FileExists(edPtr.Text) or not OpenPtr(edPtr.Text) then edPtr.Text:='';
//  if FileExists(edPtp.Text) then OpenPtp(edPtp.Text) else edPtp.Text:='';
  LastPtr:=edPtr.Text; LastTTR:=edTtr.Text;
  Reset;
  UpdateView;
  end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
  IniFile : TIniFile;
  i       : integer;
  s       : string;
begin
  Reset;
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    WriteInteger(CfgSekt,iniTop,Top);
    WriteInteger(CfgSekt,iniLeft,Left);
    WriteString(CfgSekt,IniStdPath,WorkPath);
    WriteString(CfgSekt,IniLastDmp,DumpName);
    WriteString(CfgSekt,IniLastLst,ListName);
    WriteInteger(CfgSekt,IniNumBase,integer(NumFormat));
    WriteInteger(CfgSekt,IniMemSz,MemSize);
    WriteInteger(CfgSekt,IniMemStep,edMemstep.Value);
    WriteBool(CfgSekt,IniFrontPan,itmFrontpanel.Checked);
    WriteBool(TraceSekt,IniVisible,itmShowTrace.Checked);
    WriteBool(WatchSekt,IniVisible,itmShowWatch.Checked);
    WriteBool(BpSekt,IniVisible,itmBreakpoints.Checked);
    WriteBool(CfgSekt,IniUpcase,cbUpcase.Checked);
    WriteInteger(CfgSekt,IniSwitch,FSwitches);
    WriteString(CfgSekt,IniPreSwtch,cbPreSwitches.Items.CommaText);
    WriteString(CfgSekt,IniLpt,edLpt.Text);
    WriteString(CfgSekt,IniPtp,edPtp.Text);
    if assigned(PtrFile) then s:=edPtr.Text else s:='';
    WriteString(CfgSekt,IniPtr,s);
    WriteString(CfgSekt,IniTtp,edTtp.Text);
    if assigned(TtrFile) then s:=edTtr.Text else s:='';
    WriteString(CfgSekt,IniTtr,s);
    WriteString(CfgSekt,IniTxtFile,TextFiles);
    WriteInteger(WatchSekt,IniCount,Watch.Count);
    for i:=0 to Watch.Count-1 do begin
      WriteInteger(WatchSekt,IniAddr+ZStrint(i,3),(Watch.Objects[i] as TWatchMem).Addr);
      end;
    WriteInteger(BpSekt,IniCount,Breakpoints.Count);
    for i:=0 to Breakpoints.Count-1 do begin
      WriteInteger(BpSekt,IniAddr+ZStrint(i,3),word(Breakpoints.Objects[i]));
      end;
    SaveHistory(IniFile,PtrSekt,'',true,edPtr.Items);
    SaveHistory(IniFile,TtrSekt,'',true,edTtr.Items);
    Free;
    end;
  if assigned(PtpFile) then FreeAndNil(PtpFile);
  if assigned(PtrFile) then FreeAndNil(PtrFile);
  if assigned(TtpFile) then FreeAndNil(TtpFile);
  if assigned(TtrFile) then FreeAndNil(TtrFile);
  if assigned(LptFile) then FreeAndNil(LptFile);
  end;

{------------------------------------------------------------------}
procedure TMainForm.tmViewTimer(Sender: TObject);
begin
//  edPc.Value:=PC;
  with NovaFrontpanel do if Visible then Run:=not Run;
  end;

procedure TMainForm.tswTTPunchClick(Sender: TObject);
begin
  if tswTTPunch.State=tssOn then pcDevices.ActivePage:=tsTTP;
  end;

procedure TMainForm.tswTTReaderClick(Sender: TObject);
begin
  if tswTTReader.State=tssOn then pcDevices.ActivePage:=tsTTR;
  with (IoDevices[ioTTI] as TTeletypeInput) do begin
    ReaderOn:=tswTTReader.State=tssOn;
    if not ReaderOn then tswTTReader.State:=tssOff;
    end;
  end;

procedure TMainForm.UpdateMemView;
var
  i : integer;
begin
  if Loaded then with sgMemory do begin
    RowCount:=MemSize;
//    for i:=0 to RowCount-1 do begin
    for i:=TopRow to TopRow+VisibleRowCount do begin
      Cells[1,i]:=IntToVal(i,NumFormat);
      Cells[2,i]:=IntToVal(Memory[i].Value,NumFormat);
      Cells[3,i]:=DisAssemble(i,Memory[i].Value,NumFormat,true);
      end;
    end;
  end;

procedure TMainForm.InvalidateMemView;
begin
  sgMemory.Invalidate;
  end;

procedure TMainForm.ShowRegister;
var
  i : integer;
  col : TColor;
begin
  tmView.Enabled:=FRun;
  if FRun and not btShowSteps.Down then Col:=clBtnFace
  else Col:=clWindow;
  with gbRegister do for i:=0 to ControlCount-1 do begin
    if (Controls[i] is TEdit) then (Controls[i] as TEdit).Color:=col;
    if (Controls[i] is TNumberEdit) then (Controls[i] as TNumberEdit).Color:=col;
    end;
  aiRun.Animate:=FRun;
  end;

procedure TMainForm.UpdateView;
var
  i : integer;
begin
  with gbRegister do for i:=0 to ControlCount-1 do
    if (Controls[i] is TNumberEdit) then with (Controls[i] as TNumberEdit) do begin
      if Bits>1 then NumMode:=NumFormat;  // alle auer Carry
      end;
  with edPtrByte do begin
    NumMode:=NumFormat; Invalidate;
    end;
  edPtpByte.NumMode:=NumFormat;
  UpdateMemView;
  end;

function TMainForm.ReadMemory (Addr : word) : word;
begin
  Result:=Memory[Addr].Value;
  end;

procedure TMainForm.ShowMemory (Addr : word);
begin
  sgMemory.Row:=Addr;
  end;

procedure TMainForm.rbBinClick(Sender: TObject);
begin
  NumFormat:=nmBin;
  UpdateView;
  end;

procedure TMainForm.rbOctalClick(Sender: TObject);
begin
  NumFormat:=nmOctal;
  UpdateView;
  end;

procedure TMainForm.rbDecimalClick(Sender: TObject);
begin
  NumFormat:=nmZeroDec;
  UpdateView;
  end;

procedure TMainForm.rbHexClick(Sender: TObject);
begin
  NumFormat:=nmHex;
  UpdateView;
  end;

procedure TMainForm.cbPCClick(Sender: TObject);
begin
  edPC.Bytes:=cbPC.Checked;
  end;

function TMainForm.CheckValue  (const AText : string) : boolean;
var
  n : integer;
begin
  Result:=TryOctalStrToInt(AText,n);
  end;

procedure TMainForm.itmSwitchesClick(Sender: TObject);
var
  s : string;
begin
  SelectFromListDialog.Execute(CursorPos,_('Predefined switches'),'',_('Octal values only'),
    [soEdit,soOrder],1,tcNone,defSwitches,cbPreSwitches.Items,s,true,CheckValue);
  end;

procedure TMainForm.cbPreSwitchesCloseUp(Sender: TObject);
begin
  with cbPreSwitches do begin
    Left:=170; Width:=21;
    if ItemIndex>=0 then FSwitches:=OctalStrToInt(Items[ItemIndex]);
    ItemIndex:=-1;
    ShowCmd;
    end;
  end;

procedure TMainForm.cbPreSwitchesDropDown(Sender: TObject);
begin
  with cbPreSwitches do begin
    Left:=45; Width:=156;
    end;
  end;

procedure TMainForm.cbSwitchesClick(Sender: TObject);
begin
  edSwitches.Bytes:=cbSwitches.Checked;
  end;

procedure TMainForm.cbCmdClick(Sender: TObject);
begin
  edCmd.Bytes:=cbCmd.Checked;
  end;

procedure TMainForm.cbAC0Click(Sender: TObject);
begin
  edAC0.Bytes:=cbAC0.Checked;
  end;

procedure TMainForm.cbAC1Click(Sender: TObject);
begin
  edAC1.Bytes:=cbAC1.Checked;
  end;

procedure TMainForm.cbAC2Click(Sender: TObject);
begin
  edAC2.Bytes:=cbAC2.Checked;
  end;

procedure TMainForm.cbAC3Click(Sender: TObject);
begin
  edAC3.Bytes:=cbAC3.Checked;
  end;

procedure TMainForm.itInfoClick(Sender: TObject);
begin
  InfoDialog(ProgName,ProgVersName+ProgVers+' - '+ProgVersDate+sLineBreak+VersInfo.Copyright
           +sLineBreak+'E-Mail: '+EmailAdr);
  end;

{------------------------------------------------------------------}
function TMainForm.LoadDump(const FileName : string; ClearBp : boolean) : boolean;
var
  df : TFileStream;
  i : integer;
begin
  if FileExists (FileName) then begin
    DumpName:=FileName;
    Caption:=ProgName+_(' - Memory dump: ')+ExtractFileName(DumpName); //' Speicherabbild: '
    for i:=0 to maxMemSize-1 do with Memory[i] do begin
      Value:=0; BreakPoint:=[];
      end;
    i:=0;
    df:=TFileStream.Create(FileName,fmOpenRead);
    with df do begin
      while (i<maxMemSize) and (Position<Size) do begin
        Read(Memory[i].Value,2);
        inc(i);
        end;
      Free;
      end;
    if i>MemSize then begin
      MemSize:=i;
      gbMem.Caption:=_('Program memory')+' ('+SizeToStr(MemSize)+')'; //'Programmspeicher'
      end;
    with BreakPoints do if ClearBp then begin
      Clear;
      if itmBreakpoints.Checked then BpWin.UpdateView(Count);
      end
    else begin
      for i:=0 to Count-1 do Memory[word(Objects[i])].BreakPoint:=[bpNormal];
      end;
    MemStackPtr:=0; MemStackCnt:=0;
    btLast.Enabled:=false;
    btNext.Enabled:=false;
    PC:=0;
    ShowCmd;
    UpdateWatch;
    UpdateView;
    Reset;
    Result:=true;
    end
  else begin
    ErrorDialog(_('File not found:')+sLineBreak+FileName);
    Result:=false;
    end;
  end;

procedure TMainForm.SaveDump(FileName : string);
var
  df : TFileStream;
  i : integer;
begin
  df:=TFileStream.Create(FileName,fmCreate);
  with df do begin
    for i:=0 to MemSize-1 do Write(Memory[i].Value,2);
    Free;
    end;
  end;

procedure TMainForm.itLoadDumpClick(Sender: TObject);
begin
  with OpenDialog do begin
    Title:=_('Load memory dump'); //'Speicherabbild laden';
    InitialDir:=WorkPath;
    Filter:=_('Memory dumps')+'|*.'+DmpExt+'|'+_('all')+'|*.*'; //Speicherabbild
    FileName:='';
    if Execute then LoadDump(FileName,true);
    end;
  end;

procedure TMainForm.itmReloadClick(Sender: TObject);
begin
  Reset;
  LoadDump(DumpName,false);
  end;

procedure TMainForm.itmTextfilesClick(Sender: TObject);
begin
  InputText(CursorPos,_('Text file types'),_('Comma separated list:'),false,TextFiles);
  end;

procedure TMainForm.itSaveDumpClick(Sender: TObject);
begin
  with SaveDialog do begin
    Title:=_('Save memory dump'); //'Speicherabbild sichern';
    if length(DumpName)>0 then InitialDir:=ExtractFilePath(DumpName)
    else InitialDir:=WorkPath;
//    Filename:=ExtractFileName(DumpName);
    DefaultExt:=DmpExt;
    Filter:=_('Memory dumps')+'|*.'+DmpExt+'|'+_('all')+'|*.*'; //Speicherabbild
    FileName:='';
    if Execute then begin
      Dumpname:=Filename;
      SaveDump(FileName);
      Caption:=ProgName+_(' - Memory dump: ')+ExtractFileName(DumpName); //' Speicherabbild: '
      end;
    end;
  end;

procedure TMainForm.itmMemResetClick(Sender: TObject);
var
  i : integer;
begin
  Caption:=ProgName;
  for i:=0 to MemSize-1 do Memory[i].Value:=0;
  PC:=0;
  UpdateView;
  ShowCmd;
  end;

{------------------------------------------------------------------}
procedure TMainForm.itmWorkDirClick(Sender: TObject);
begin
  ShellDirDialog.Execute (_('Work directory'),false,true,false,'',WorkPath); //'Arbeitsverzeichnis'
  end;

procedure TMainForm.itEndeClick(Sender: TObject);
begin
  Close;
  end;

{------------------------------------------------------------------}
procedure TMainForm.ClearRegs;
begin
  PC:=0; FCarry:=0;
  Akkus[0]:=0; Akkus[1]:=0; Akkus[2]:=0; Akkus[3]:=0;
  end;

function TMainForm.IntAck : word;
var
  i : integer;
begin
  i:=0;
  while (i<NumberOfDevices) and not IoDevices[DeviceChain[i]].Interrupt do inc(i);
  if i<NumberOfDevices then Result:=DeviceChain[i]
  else Result:=0;
  end;

procedure TMainForm.MaskOut(Value : word);
var
  i : integer;
begin
  for i:=FirstDevice to LastDevice do
    if assigned(IoDevices[i]) then IoDevices[i].MaskOut(Value);
  end;

procedure TMainForm.mmTTOKeyPress(Sender: TObject; var Key: Char);
begin
  Key:=#0;
  end;

procedure TMainForm.IoReset;
var
  i : integer;
begin
  for i:=FirstDevice to LastDevice do if assigned(IoDevices[i]) then IoDevices[i].Reset;
  end;

procedure TMainForm.Reset;
begin
  FRun:=false;
  Trace.Clear;
  TraceWin.UpdateView(0);
  btStop.Enabled:=FRun;
  PreEnable:=false; IntEnable:=false; IrqFlag:=false;
  IoReset;
  ClearRegs;
  PC:=StartAdr;
  ShowCmd;
  end;

{------------------------------------------------------------------}
procedure TMainForm.AddBreakpoint (AAddr : word);
begin
  Breakpoints.AddObject(IntToOctal(AAddr,6,0),pointer(AAddr));
  end;

procedure TMainForm.RemBreakpoint (AAddr : word);
var
  n : integer;
begin
  with Breakpoints do begin
    n:=IndexOf(IntToOctal(AAddr,6,0));
    if n>=0 then Delete(n);
    end;
  end;

{------------------------------------------------------------------}
procedure TMainForm.AddToWatch (AAddr,AData : word);
begin
  Watch.AddObject(IntToOctal(AAddr,6,0),TWatchMem.Create(AAddr,AData));
  if assigned(WatchWin) then with WatchWin do if Visible then UpdateView(Watch.Count);
  end;

function TMainForm.UpdateWatch (WatchChanges : boolean) : boolean;
var
  i : integer;
  ch : boolean;
begin
  Result:=false; ch:=false;
  with Watch do for i:=0 to Count-1 do with Objects[i] as TWatchMem do begin
    Changed:=WatchChanges and (Data<>Memory[Addr].Value);
    ch:=ch or Changed;
    Data:=Memory[Addr].Value;
    Result:=Result or (Alert and Changed);
    end;
  with WatchWin do if Visible and (not WatchChanges or ch) then UpdateView(Watch.Count);
  end;

procedure TMainForm.ClearWatches;
begin
  FreeListObjects(Watch); Watch.Clear;
  end;

function TMainForm.DeleteWatch (const sAddr : string) : integer;
var
  n : integer;
begin
  with Watch do begin
    n:=IndexOf(sAddr);
    if n>=0 then begin
      Objects[n].Free; Delete(n);
      end;
    Result:=Count;
    end;
  end;

{------------------------------------------------------------------}
procedure TMainForm.ShowCmd;
begin
  siIntEn.Indicate:=IntEnable; siIrq.Indicate:=IrqFlag;
  with sgMemory do begin
    if PC<=RowCount then Row:=PC
    else ErrorDialog(_('Address out of range:'+sLineBreak+'0 .. ')+IntToVal(MemSize,NumFormat));
      // 'Adresse auerhalb des Speicherbereichs:'#13+'0 .. '
    if (Row>1) and (Row-TopRow>VisibleRowCount-2) then TopRow:=Row-1;
    end;
  edPC.Value:=PC;
  edCMD.Value:=Memory[PC].Value;
  edCommand.Text:=DisAssemble(PC,edCMD.Value,NumFormat,false);
  edCarry.Value:=FCarry;
  edAC0.Value:=Akkus[0];
  edAC1.Value:=Akkus[1];
  edAC2.Value:=Akkus[2];
  edAC3.Value:=Akkus[3];
  edAdr.Text:='';
  with aiSkip do begin
    Indicate:=AnalyseCmd(edCMD.Value,false);
    if Indicate then Direction:=adDown else Direction:=adRight;
    end;
  edSwitches.Value:=FSwitches;
  siRun.Indicate:=FRun;
  with NovaFrontpanel do if Visible then begin
    Run:=FRun;
    Address:=PC;
    Carry:=FCarry<>0;
    Data:=edCMD.Value;
    Switches:=FSwitches;
    Application.ProcessMessages;
    end;
  end;

// execute inst., return true on skip
function TMainForm.AnalyseCmd(cmd : word; Execute : boolean) : boolean;
var
  shreg     : TLongWord;
  OldPc,fct,
  acs,nacd,
  cry,inst  : word;
  PHalt     : boolean;
  i         : integer;
begin
  Result:=false; PHalt:=false;
  OldPc:=PC;
  if Execute then begin
    for i:=0 to NumberOfDevices-1 do IoDevices[DeviceChain[i]].ProcessTick;
    end;
  nacd:=(Hi(cmd) and $18) shr 3;
  if (cmd and $8000)<>0 then begin  // execute arithm inst.
    acs:=Akkus[(Hi(cmd) and $60) shr 5];          // source ac
    case (cmd and $30) shr 4 of                   // carry condition
    1 : cry:=0;                                   // Z
    2 : cry:=1;                                   // O
    3 : cry:=FCarry xor 1;                        // C
    else cry:=FCarry;                             // unchanged
      end;
    inst:=Hi(cmd and $300);                       // function
    with shreg do begin
      if cmd and $400 =0 then begin               // single operand
        Hi:=cry; Lo:=0;
        case inst of
        0 : LongWord:=LongWord+not acs;           // COM
        1 : LongWord:=LongWord+not acs+1;         // NEG
        2 : LongWord:=LongWord+acs;               // MOV
        3 : LongWord:=LongWord+acs+1;             // INC
          end;
        end
      else begin
        Hi:=cry; Lo:=Akkus[nacd];                 // dual operand
        case inst of
        0 : LongWord:=LongWord+not acs;           // ADC
        1 : LongWord:=LongWord+not acs+1;         // SUB
        2 : LongWord:=LongWord+acs;               // ADD
        3 : Lo:=Lo and acs;                       // AND
          end;
        end;
      LongWord:=LongWord and $1FFFF;              // shift condition
      inst:=(cmd and $C0) shr 6;
      case inst of
      1 : LongWord:=(LongWord shl 1) and $1FFFE or Hi;    // L
      2 : begin
          Hi:=Hi or (Lo and 1) shl 1;                     // R
          LongWord:=LongWord shr 1;
          end;
      3 : begin
          cry:=LoH; LoH:=LoL; LoL:=cry;                   // S
          end;
        end;
      if Execute and (cmd and 8=0) then begin    // noload conditon
        Akkus[nacd]:=Lo;
        FCarry:=Hi;
        end;
      inst:=cmd and 7;                           // skip condition
      case inst of
      1 : Result:=true;                                    // SKP
      2 : Result:=Hi and 1 =0;                             // SZC
      3 : Result:=Hi and 1 <>0;                            // SNC
      4 : Result:=Lo and $FFFF =0;                         // SZR
      5 : Result:=Lo and $FFFF <>0;                        // SNR
      6 : Result:=(Hi and 1 =0) xor (Lo and $FFFF =0);     // SEZ
      7 : Result:=(Hi and 1 <>0) and (Lo and $FFFF <>0);   // SBN
        end;
      end;
    if Execute then inc(PC);
    end
  else begin
    if cmd and $6000 = 0 then begin   // jmp, jsr, isz, dsz
      inst:=ResolveAdr(cmd,Execute);
      case nacd of
      0 : if Execute then PC:=inst;   // jmp
      1 : if Execute then begin       // jsr
          Akkus[3]:=PC+1;             // return addr.
          PC:=inst;
          end;
      2 : begin                       // isz
          acs:=Memory[inst].Value+1;
          Result:=acs=0;
          if Execute then begin
            Memory[inst].Value:=acs;
            inc(PC);
            end;
          end;
      3 : begin                       // dsz
          acs:=Memory[inst].Value-1;
          Result:=acs=0;
          if Execute then begin
            Memory[inst].Value:=acs;
            inc(PC);
            end;
          end;
        end;
      end
    else if cmd and $6000 = $6000 then begin // In-Out
      if cmd=UIMUL then with shreg do begin
        LongWord:=Akkus[2];
        LongWord:=LongWord*Akkus[1]+Akkus[0];
        if Execute then begin
          Akkus[0]:=Hi; Akkus[1]:=Lo;
          end;
        end
      else if cmd=UIDIV then with shreg do begin
        Hi:=Akkus[0];
        if Hi>=Akkus[2] then FCarry:=1
        else begin
          FCarry:=0;
          Lo:=Akkus[1];
          if Execute then DivMod(LongWord,Akkus[2],Akkus[1],Akkus[0]);
          end;
        end
      else begin
        inst:=cmd and $3F;  // Dev. addr.
        fct:=Hi(cmd) and 7;
        acs:=(cmd shr 6) and 3;
        if inst=ioCPU then begin // CPU commands
          if fct=7 then begin // SKP
            case acs of
            0 : Result:=IntEnable;
            1 : Result:=not IntEnable;
            2 : Result:=false; // Powerfail
            3 : Result:=true;
              end;
            end
          else if Execute then begin
            case fct of
            1 : Akkus[nacd]:=FSwitches;
            3 : Akkus[nacd]:=IntAck;        // INTA
            4 : MaskOut(Akkus[nacd]);       // MSKO
            5 : IoReset;
            6 : PHalt:=true;                // Halt
              end;
            case acs of
            1 : PreEnable:=true;
            2 : begin
                IntEnable:=false;
                PreEnable:=false;
                end;
              end;
            end;
          end
        else if (inst>=FirstDevice) and (IoDevices[inst]<>nil) then begin
          with IoDevices[inst] do begin  // known device
            if fct=7 then begin // SKP
              case acs of
              0 : Result:=BusyFlag;
              1 : Result:=not BusyFlag;
              2 : Result:=DoneFlag;
              3 : Result:=not DoneFlag;
                end;
              end
            else if Execute then begin
              case fct of
              1 : Akkus[nacd]:=DIA;
              2 : DOA(Akkus[nacd]);
              3 : Akkus[nacd]:=DIB;
              4 : DOB(Akkus[nacd]);
              5 : Akkus[nacd]:=DIC;
              6 : DOC(Akkus[nacd]);
                end;
              case acs of
              1 : Start;
              2 : Clear;
              3 : Pulse;
                end;
              end;
            end;
          end
        else begin // unknown device
          if fct=7 then begin // SKP
            case acs of
            0 : Result:=false;
            1 : Result:=true;
            2 : Result:=false;
            3 : Result:=true;
              end;
            end
          else if Execute then begin
            case fct of
            1 : Akkus[nacd]:=0;
            3 : Akkus[nacd]:=0;
            5 : Akkus[nacd]:=0;
              end;
            end;
          end;
      end;
      if Execute and not PHalt then inc(PC);
      if PHalt then FRun:=false;
      end
    else begin  // lda, sta
      inst:=ResolveAdr(cmd,Execute);
      if Execute then begin
        if cmd and $6000 = $2000 then Akkus[nacd]:=Memory[inst].Value
        else Memory[inst].Value:=Akkus[nacd];
        inc(PC);
        end;
      Result:=false;
      end;
    end;
  if Execute then begin
    if UpdateWatch(true) then FRun:=false; // stop
    AddToTrace(TTraceData.Create(OldPC,cmd,FCarry,Akkus));
    end;
  end;

procedure TMainForm.AddToTrace (TrData : TTraceData);
begin
  with Trace do begin
    if Count>=MaxTraceItems then Delete(0);
    Add(TrData);
    TraceWin.UpdateView(Count);
    end;
  end;

function TMainForm.IndirectAdr (Addr : word; Execute : boolean) : word; // ind. Addr. auflsen
begin
  repeat
    Addr:=Addr and $7FFF;
    if Execute and (Addr>=$10) and (Addr<=$1F) then begin
      if (Addr<=$17) then inc(Memory[Addr].Value) else dec(Memory[Addr].Value);
      end;
    Addr:=Memory[Addr].Value;
    until Addr and $8000 = 0;
  Result:=Addr;
  end;


function TMainForm.ResolveAdr(Command : word; Execute : boolean) : word;
var
  ind  : boolean;
  ndx  : word;
  disp : smallint;
begin
  ind:=Command and $400 <>0;
  ndx:=Hi(Command) and 3;
  disp:=Lo(Command);
  if disp>127 then disp:=disp-256;
  case ndx of
  0 : if ind then Result:=IndirectAdr(Lo(Command),Execute) else Result:=Lo(Command);
  1 : if ind then Result:=IndirectAdr(PC+disp,Execute) else Result:=PC+disp;
  2 : if ind then Result:=IndirectAdr(Akkus[2]+disp,Execute) else Result:=Akkus[2]+disp;
  3 : if ind then Result:=IndirectAdr(Akkus[3]+disp,Execute) else Result:=Akkus[3]+disp;
    end;
  if not Execute then edAdr.Value:=Result;
  end;

procedure TMainForm.itmMemSizeClick(Sender: TObject);
var
  n : word;
begin
  n:=MemSize;
  if ValueDialog.Execute(CursorPos,_('Memory size in words:'),NumFormat,n) then begin //'Speichergre in Worten:'
    if (n>=MemSize) or
      ConfirmDialog(CursorPos,Format(_('Do you really want to reduce memory size to'+sLineBreak+'%s words?'),[IntToVal(n,NumFormat)]))
      // Soll die Speichergre wirklich auf'#13+'%s+' Worte verkleinert werden?',CursorPos)
        then begin
      MemSize:=n;
      gbMem.Caption:=_('Program memory')+' ('+SizeToStr(MemSize)+')'; //'Programmspeicher'
      end;
    sgMemory.RowCount:=MemSize+1;
    UpdateView;
    end;
  end;

procedure TMainForm.edPCClick(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edPC),_('Program counter:'),NumFormat,PC) then begin //'Programmzhler (PC):'
    ShowCmd;
    sgMemory.Invalidate;
    end;
  end;

procedure TMainForm.edCarryClick(Sender: TObject);
var
  v : word;
begin
  v:=FCarry;
  if ValueDialog.Execute(BottomLeftPos(edCarry),_('Carry:'),NumFormat,v) then begin //'Carry:'
    FCarry:=v and 1;
    ShowCmd;
    edAc0.SetFocus;
    end;
  end;

procedure TMainForm.edAC0Click(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edAC0),_('Accumulator 0:'),NumFormat,Akkus[0]) then begin
    ShowCmd; //'Akkumulator 0:'
    edAc1.SetFocus;
    end;
  end;

procedure TMainForm.edAC1Click(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edAC1),_('Accumulator 1:'),NumFormat,Akkus[1]) then begin
    ShowCmd;
    edAc2.SetFocus;
    end;
  end;

procedure TMainForm.edAC2Click(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edAC2),_('Accumulator 2:'),NumFormat,Akkus[2]) then begin
    ShowCmd;
    edAc3.SetFocus;
    end;
  end;

procedure TMainForm.edAC3Click(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edAC3),_('Accumulator 3:'),NumFormat,Akkus[3]) then begin
    ShowCmd;
    edAc0.SetFocus;
    end;
  end;

procedure TMainForm.edSwitchesClick(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edSwitches),_('Front panel switches:'),NumFormat,FSwitches) then
    ShowCmd; //'Frontplatten-Schalter:'
  end;

procedure TMainForm.DataSwitchChanged (Value : word);
begin
  FSwitches:=Value;
  edSwitches.Value:=FSwitches;
//  ShowCmd;
  end;

procedure TMainForm.CommandSwitchChanged (Sender : TObject; Command : TNovaCommands);
begin
  with (Sender as TNovaFrontPanel) do case Command of
  ncDepAc0  : begin
      Akkus[0]:=Switches;
      edAC0.Value:=Akkus[0];
      Data:=Akkus[0];
      end;
  ncExAc0  : Data:=Akkus[0];
  ncDepAc1 : begin
      Akkus[1]:=Switches;
      edAC1.Value:=Akkus[1];
      Data:=Akkus[1];
      end;
  ncExAc1  : Data:=Akkus[1];
  ncDepAc2 :  begin
      Akkus[2]:=Switches;
      edAC2.Value:=Akkus[2];
      Data:=Akkus[2];
      end;
  ncExAc2  : Data:=Akkus[2];
  ncDepAc3 : begin
      Akkus[3]:=Switches;
      edAC3.Value:=Akkus[3];
      Data:=Akkus[3];
      end;
  ncExAc3 : Data:=Akkus[3];
  ncReset : Reset;
  ncStop  : btStopClick(self);
  ncStart : begin
      PC:=Switches and $7FFF;
      edPC.Value:=PC;
      RunProg;
      end;
  ncCont : RunProg;
  ncDep  : begin
      Memory[Address].Value:=Switches;
      Data:=Memory[Address].Value;
      UpdateMemView;
      end;
  ncDepNext : begin
      Address:=(Address+1) and $7FFF;
      Memory[Address].Value:=Switches;
      Data:=Memory[Address].Value;
      UpdateMemView;
      end;
  ncEx : begin
      Address:=Switches and $7FFF;
      Data:=Memory[Address].Value;
      end;
  ncExNext : begin
      Address:=(Address+1) and $7FFF;
      Data:=Memory[Address].Value;
      end;
  ncMemStep : ;
  ncInstStep : btStepClick(self);
  ncProgLoad : ProgramLoad;
    end;
  end;

procedure TMainForm.edCmdClick(Sender: TObject);
begin
  if CommandDialog.Execute(BottomLeftPos(edCmd),_('Assembler command:'),PC,
      NumFormat,Memory[PC].Value) then begin //'Assembler-Befehl:'
    with sgMemory do begin
      Cells[2,PC]:=IntToVal(Memory[PC].Value,NumFormat);
      Cells[3,PC]:=DisAssemble(PC,Memory[PC].Value,NumFormat,true);
      end;
    ShowCmd;
    end;
  end;

procedure TMainForm.GotoMem (Adr : integer);
begin
  with sgMemory do begin
    if Adr>=0 then begin
      if Adr<RowCount then Row:=Adr else Row:=RowCount-1;
      end
    else Row:=0;
    if Row<VisibleRowCount div 2 then TopRow:=0
    else begin
      if RowCount-Row<VisibleRowCount div 2 then TopRow:=RowCount-VisibleRowCount
      else TopRow:=Row-VisibleRowCount div 2;
      end;
    Invalidate;
    end;
  end;

procedure TMainForm.SaveMemPos (Adr : integer);
var
  i : integer;
begin
  if (MemStack[MemStackPtr]<>Adr) then begin
    if MemStackPtr>=MemStackSize-1 then
      for i:=1 to MemStackSize-1 do MemStack[i-1]:=MemStack[i];
    MemStack[MemStackPtr]:=Adr;
    inc(MemStackPtr);
    MemStackCnt:=MemStackPtr;
    btLast.Enabled:=MemStackCnt>1;
    btNext.Enabled:=false;
    end;
  end;

procedure TMainForm.btLastClick(Sender: TObject);
begin
  if MemStackPtr>0 then begin
    if MemStackPtr>=MemStackCnt then dec(MemStackPtr);
    dec(MemStackPtr);
    GotoMem(MemStack[MemStackPtr]);
    end;
  btLast.Enabled:=MemStackPtr>0;
  btNext.Enabled:=true;
  end;

procedure TMainForm.btNextClick(Sender: TObject);
begin
  if MemStackPtr<MemStackCnt-1 then begin
    inc(MemStackPtr);
    GotoMem(MemStack[MemStackPtr]);
    end;
  btNext.Enabled:=MemStackPtr<MemStackCnt-1;
  btLast.Enabled:=MemStackPtr>0;
  end;

procedure TMainForm.sgMemoryMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ac,ar : integer;
  cmd   : word;
  pt    : TPoint;
begin
  with sgMemory do begin
    MouseToCell(x,y,ac,ar);
    SaveMemPos(Row);
    Row:=ar;
    Invalidate;
    cmd:=Memory[Row].Value;
    ContentsCol:=ac=2;
    itpSeek.Enabled:=(ac=2) or (cmd and $E000<$6000) and (Hi(cmd) and 2=0);
    pt:=ClientToScreen(Point(x,y+RowHeights[Row]));
    end;
  if Button=mbRight then with pt do pmMemory.Popup(x,y);
  end;

procedure TMainForm.sgMemoryDblClick(Sender: TObject);
begin
  with sgMemory do begin
    if Col=0 then with Memory[Row] do begin
      if bpNormal in BreakPoint then DelBreakpoint(sgMemory.Row)
      else SetBreakpoint(sgMemory.Row);
      end
    else if Col=2 then begin
      if ValueDialog.Execute(CursorPos,Format(_('Memory cell %s - Value:'),[IntToVal(Row,NumFormat)]),NumFormat,Memory[Row].Value)
      // 'Speicherzelle %s - Wert:'
          then begin
        Cells[2,Row]:=IntToVal(Memory[Row].Value,NumFormat);
        Cells[3,Row]:=DisAssemble(Row,Memory[Row].Value,NumFormat,true);
        if Row=PC then ShowCmd;
        end;
      end
    else if Col=3 then begin
      if CommandDialog.Execute(CursorPos,Format(_('Memory cell %s - Command:'),[IntToVal(Row,NumFormat)]),
      //'Speicherzelle %s - Befehl:',
                 Row,NumFormat,Memory[Row].Value) then begin
        with sgMemory do begin
          Cells[2,Row]:=IntToVal(Memory[Row].Value,NumFormat);
          Cells[3,Row]:=DisAssemble(Row,Memory[Row].Value,NumFormat,true);
          end;
        if Row=PC then ShowCmd;
        end;
      end;
    Invalidate;
    end;
  end;

procedure TMainForm.sgMemoryKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_RETURN then begin
    sgMemoryDblClick(Sender);
    with sgMemory do if RowCount-1>Row then Row:=Row+1;
    end;
  end;

procedure TMainForm.sgMemoryDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
begin
  with sgMemory,Canvas do begin
    if FRun and not btShowSteps.Down then Brush.Color:=clBtnFace
    else if ARow=PC then Brush.Color:=$9D9DFF
    else if ARow=Row then Brush.Color:=$D9C4BB
    else if gdSelected in State then Brush.Color:=$A2E8BF
    else Brush.Color:=clWindow;
    FillRect(Rect);
    if (ACol=0) then begin
      Brush.Color:=clRed;
      with Pen do begin
        Width:=1;
        Color:=clBlack
        end;
      if bpNormal in Memory[ARow].BreakPoint then
        with Rect do Ellipse(Left+2,Top+2,Left+Bottom-Top-2,Bottom-2);
      end
    else TextOut(Rect.Left+2,Rect.Top,Cells[ACol,ARow]);
    end;
  end;

procedure TMainForm.sgMemoryTopLeftChanged(Sender: TObject);
begin
  UpdateMemView;
  end;

procedure TMainForm.btResetClick(Sender: TObject);
begin
  Reset;
  with mmTTO.Lines do begin
    Clear;
    Add('');
    end;
  end;

procedure TMainForm.btShowStepsClick(Sender: TObject);
begin
  ShowRegister;
  sgMemory.Invalidate;
  end;

procedure TMainForm.btStepClick(Sender: TObject);
begin
  if AnalyseCmd(Memory[PC].Value,true) then inc(PC);
  ProcessInterrupts;
  UpdateView;
  ShowCmd;
  end;

procedure TMainForm.btStepOverClick(Sender: TObject);
begin
  if (PC=MemSize-1) or ((Memory[PC].Value and $F800) <>$800) then btStepClick(Sender)
  else begin   //JSR
    Include(Memory[PC+1].BreakPoint,bpStepOver);
    RunProg;
    end;
  end;

procedure TMainForm.btRunClick(Sender: TObject);
begin
  RunProg;
  end;

procedure TMainForm.RunProg;
begin
  if Memory[PC].Value=PHalt then Inc(PC);
  FRun:=true;
  sgMemory.Invalidate;
  ShowRegister;
  with NovaFrontPanel do if Visible then Run:=FRun;
  btStop.Enabled:=FRun;
  siRun.Indicate:=FRun;
  while FRun do begin
    if AnalyseCmd(Memory[PC].Value,true) then inc(PC);   // Skip
    ProcessInterrupts;
    if btShowSteps.Down then ShowCmd;
    Application.ProcessMessages;
    if FRun then with Memory[PC] do begin
      FRun:=not (bpNormal in BreakPoint);
      if FRun and (bpStepOver in BreakPoint) then begin
        FRun:=false;
        Exclude(Breakpoint,bpStepOver);
        end;
      end;
    end;
  btStop.Enabled:=FRun;
  sgMemory.Invalidate;
  ShowRegister;
  UpdateView;
  ShowCmd;
  end;

procedure TMainForm.btStopClick(Sender: TObject);
begin
  FRun:=false;
  btStop.Enabled:=FRun;
  ShowRegister;
  with NovaFrontPanel do if Visible then Run:=FRun;
  siRun.Indicate:=FRun;
  end;

procedure TMainForm.ProcessInterrupts;
begin
  if IrqFlag and IntEnable then begin
    PreEnable:=false;
    IntEnable:=false;
    IrqFlag:=false;
    Memory[0].Value:=PC;
    PC:=IndirectAdr(1,true);
    end;
  IntEnable:=PreEnable;
  end;

procedure TMainForm.SetInterrupt (Sender: TObject);
begin
  IrqFlag:=true;
  end;

//-----------------------------------------------------------------------------
function TMainForm.OpenOutputFile (ATitle,AFilter,AExt : string; var AFilename : string) : boolean;
begin
  with SaveDialog do begin
    Title:=ATitle;
    if length(AFileName)>0 then InitialDir:=ExtractFilePath(AFileName)
    else InitialDir:=WorkPath;
    Filename:=ExtractFileName(AFileName);
    DefaultExt:=AExt;
    Filter:=AFilter;
    Result:=Execute;
    if Result then AFilename:=Filename;
    end;
  end;

function TMainForm.OpenInputFile (ATitle,AFilter,AExt : string; var AFilename : string) : boolean;
begin
  with OpenDialog do begin
    Title:=ATitle;
    if length(AFileName)>0 then InitialDir:=ExtractFilePath(AFileName)
    else InitialDir:=WorkPath;
    Filename:=ExtractFileName(AFileName);
    DefaultExt:=AExt;
    Filter:=AFilter;
    FileName:='';
    Result:=Execute;
    if Result then AFilename:=Filename;
    end;
  end;

function TMainForm.SetTextFileFilter (const ExtList : string) : string;
var
  s : string;
begin
  s:=ExtList;
  Result:=_('Text files')+'|*.'+ReadNxtStr(s,',');
  repeat
    Result:=Result+';*.'+ReadNxtStr(s,',');
    until length(s)=0;
  Result:=Result+'|';
  end;

//-----------------------------------------------------------------------------
function TMainForm.OpenTtr (const Filename : string) : boolean;
begin
  Result:=false;
  try
    TtrFile:=TInputStream.Create(Filename);
    edTtrTotal.Value:=TtrFile.Size;
    edTtrPos.Value:=0; edTtrByte.Value:=0;
    (IoDevices[ioTti] as TInDevice).InputByte:=0;
    if HasTextExt(Filename) then rbTtrTextFile.Checked:=true else rbTtrBinFile.Checked:=true;
    Result:=true;
  except
    try FreeAndNil(TtrFile); except end;
    end;
  if not Result then ErrorDialog(_('Error opening file:')+sLineBreak+Filename);
  end;

function TMainForm.SelectTtrFile : boolean;
var
  s     : string;
begin
  s:=edTtr.Text; Result:=false;
  if OpenInputFile(_('TTY reader input from file'),
      SetTextFileFilter(TextFiles)+_('Binary files')+'|*.bin|'+_('all')+'|*.*','',s) then begin
    if OpenTtr(s) then begin
      AddToHistory(edTtr.Items,s);
      edTtr.ItemIndex:=0;
      Result:=true;
      end
    else edTtr.Text:='';
    end
  end;

procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  with (IoDevices[ioTTI] as TTeletypeInput) do if cbUpcase.Checked then InputByte:=ord(Upcase(Key))
  else InputByte:=ord(Key);
  end;

procedure TMainForm.ReadFromTtr (Sender: TObject);
var
  AByte : byte;
  eof   : boolean;
begin
  if not assigned(TtrFile) then begin
    if not SelectTtrFile then begin
      btStopClick(Sender);
      (Sender as TInDevice).Reset;
      Exit;
      end;
    end;
  if assigned(TtrFile) then with TtrFile do begin
    eof:=Position>=Size;
    if not eof then Read(AByte,1)
    else if TrailerCount<TrailerLength then begin
      if rbPtrTextFile.Checked and (TrailerCount=0) then AByte:=$0C  // one form feed
      else AByte:=0;
      inc(TrailerCount);
      end;
    if TrailerCount<TrailerLength then (Sender as TInDevice).InputByte:=AByte;
    if (Position mod 10 =0) or eof then begin
      edTtrPos.Value:=Position;
      edTtrByte.Value:=AByte;
      end;
    end;
  end;

procedure TMainForm.edTTRCloseUp(Sender: TObject);
var
  ok : boolean;
  s  : string;
begin
  with edTTR do s:=Items[ItemIndex];
  if length(s)=0 then Exit;
  if assigned(TTRFile) then begin
    ok:=(TtrFile.TrailerCount>=TrailerLength)
      or ConfirmDialog(TopRightPos(btTtrOpen),Format(_('Close TTY reader file:'+sLineBreak+'%s?'),[TTRFile.FName]));
    if ok then FreeAndNil(TTRFile);
    end;
  if not assigned(TTRFile) then begin
    if FileExists(s) then ok:=OpenTTR(s)
    else begin
      ErrorDialog(_('File not found:')+sLineBreak+s);
      ok:=false;
      end;
    with edTTR do if ok then Invalidate
    else ItemIndex:=-1;
    end
  else with edTTR do begin
    ItemIndex:=Items.IndexOf(LastTTR);
    end;
  end;

procedure TMainForm.edTTRDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
begin
  with edTTR do begin
    with Canvas do begin
      if (odSelected in State) and not (odComboBoxEdit in State) then Brush.Color:=clHighlight
      else Brush.Color:=clWindow;
      if (odComboBoxEdit in State) and assigned(TTRFile)  then Font.Color:=clWindowText;
      FillRect(Rect);
      TextOut(Rect.Left+2,Rect.Top,Items[Index]);
      end;
    end;
  end;

procedure TMainForm.edTtrCloseClick(Sender: TObject);
begin
  if assigned(PtrFile) then PtrFile.TrailerCount:=TrailerLength; // set to end
//  FreeAndNil(TtrFile);
  edTtr.ItemIndex:=-1;
  end;

procedure TMainForm.edTtrDropDown(Sender: TObject);
begin
  LastTTR:=edTTR.Text;
  end;

procedure TMainForm.btTtrOpenClick(Sender: TObject);
var
  ok : boolean;
begin
  if assigned(TtrFile) then begin
    ok:=(TtrFile.TrailerCount>=TrailerLength)
      or ConfirmDialog(TopRightPos(btTtrOpen),Format(_('Close TTY reader file:'+sLineBreak+'%s?'),[TtrFile.FName]));
    if ok then FreeAndNil(TtrFile);
    end
  else ok:=true;
  if ok then ok:=SelectTtrFile;
  if not ok then edTtr.ItemIndex:=-1;
  end;

procedure TMainForm.edTtrResetClick(Sender: TObject);
begin
  if assigned(TtrFile) then with TtrFile do begin
    Position:=0; TtrFile.TrailerCount:=0;
    edTtrPos.Value:=0; edTtrByte.Value:=0;
    (IoDevices[ioTTI] as TInDevice).InputByte:=0;
    end;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.WriteToTTO (Sender: TObject; AByte : byte);
begin
  AByte:=AByte and $7F;   // mask out parity bit
  if tswTTPunch.State=tssOn then begin
    if not assigned(TtpFile) then begin
      if not SelectTtpFile then begin
        btStopClick(Sender); Exit;
        end;
      end;
    if assigned(TtpFile) then with TtpFile do begin
      Write(AByte,1);
      if (Position mod 10 =0) or (AByte=$C) then begin
        edTtpByte.Value:=AByte;
        edTtpPos.Value:=Position;
        end;
      end;
    end;
  with mmTTO.Lines do begin
    if AByte=13 then Add('>')
    else if AByte>=32 then Strings[Count-1]:=Strings[Count-1]+chr(AByte);
    Application.ProcessMessages;
    end;
  end;

function TMainForm.OpenTtp (const Filename : string) : boolean;
begin
  Result:=false;
  try
    TtpFile:=TFileStream.Create(Filename,fmCreate+fmShareDenyWrite);
    edTtpPos.Value:=0;
    Result:=true;
  except
    try FreeAndNil(PtrFile); except end;
    end;
  end;

function TMainForm.SelectTtpFile : boolean;
var
  s : string;
begin
  s:=edTtp.Text; Result:=false;
  if OpenOutputFile(_('TTY punch output to file'),_('all')+'|*.*','bin',s) then begin
    if OpenTtp(s) then edTtp.Text:=s
    else begin
      ErrorDialog(_('Error opening file:')+sLineBreak+s);
      edTtp.Text:='';
      Result:=true;
      end;
    end
  end;

procedure TMainForm.btTtpOpenClick(Sender: TObject);
begin
  if assigned(TtpFile) and
      ConfirmDialog(TopRightPos(btTtpOpen),Format(_('Close TTY punch output file:'+sLineBreak+'%s?'),[edTtp.Text])) then
    FreeAndNil(TtpFile);
  if not assigned(TtpFile) then SelectTtpFile;
  end;

procedure TMainForm.edTtpCloseClick(Sender: TObject);
begin
  if assigned(TtpFile) and
      ConfirmDialog(TopRightPos(btTtpOpen),Format(_('Close TTY punch output file:'+sLineBreak+'%s?'),[edTtp.Text])) then begin
    FreeAndNil(TtpFile);
    edTtp.Text:='';
    edTtpPos.Value:=0;
    end;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.tmRtcTimer(Sender: TObject);
begin
  (IoDevices[ioRtc] as TRtcDevice).TimeElapsed;
  end;

//-----------------------------------------------------------------------------
function TMainForm.OpenPtp (const Filename : string) : boolean;
begin
  Result:=false;
  try
    PtpFile:=TFileStream.Create(Filename,fmCreate+fmShareDenyWrite);
    edPtpPos.Value:=0;
    Result:=true;
  except
    try FreeAndNil(PtrFile); except end;
    end;
  end;

function TMainForm.SelectPtpFile : boolean;
var
  s : string;
begin
  s:=edPtp.Text; Result:=false;
  if OpenOutputFile(_('PTP output to file'),_('all')+'|*.*','bin',s) then begin
    if OpenPtp(s) then edPtp.Text:=s
    else begin
      ErrorDialog(_('Error opening file:')+sLineBreak+s);
      edPtp.Text:='';
      Result:=true;
      end;
    end
  end;

procedure TMainForm.WriteToPTP (Sender: TObject; AByte : byte);
begin
  if not assigned(PtpFile) then begin
    if not SelectPtpFile then begin
      btStopClick(Sender); Exit;
      end;
    end;
  if assigned(PtpFile) then with PtpFile do begin
    Write(AByte,1);
    if (Position mod 10 =0) or (AByte=$C) then begin
      edPtpByte.Value:=AByte;
      edPtpPos.Value:=Position;
      end;
    end;
  end;

procedure TMainForm.btPtpCloseClick(Sender: TObject);
begin
  if assigned(PtpFile) and
      ConfirmDialog(TopRightPos(btPtpOpen),Format(_('Close PTP output file:'+sLineBreak+'%s?'),[edPtp.Text])) then begin
    FreeAndNil(PtpFile);
    edPtp.Text:='';
    edPtpPos.Value:=0;
    end;
  end;

procedure TMainForm.btPtpOpenClick(Sender: TObject);
begin
  if assigned(PtpFile) and
      ConfirmDialog(TopRightPos(btPtpOpen),Format(_('Close PTP output file:'+sLineBreak+'%s?'),[edPtp.Text])) then
    FreeAndNil(PtpFile);
  if not assigned(PtpFile) then SelectPtpFile;
  end;

//-----------------------------------------------------------------------------
function TMainForm.HasTextExt (const Filename : string) : boolean;
var
  se,s : string;
begin
  se:=GetExt(Filename); s:=TextFiles;
  repeat
    Result:=AnsiSameText(se,ReadNxtStr(s,','));
    until Result or (length(s)=0);
  end;

function TMainForm.NoMoreBytes (fs : TStream) : boolean;
var
  sp,i : int64;
  b    : byte;
begin
  with fs do begin             // check if only zero byte left
    sp:=Position; Result:=true;
    i:=sp;
    while i<Size do begin
      Read(b,1);
      Result:=b=0;
      if not Result then Break;
      inc(i);
      end;
    Position:=sp;
    end;
  end;

function TMainForm.OpenPtr (const Filename : string) : boolean;
begin
  Result:=false;
  try
    PtrFile:=TInputStream.Create(Filename);
    edPtrTotal.Value:=PtrFile.Size;
    edPtrPos.Value:=0; edPtrByte.Value:=0;
    (IoDevices[ioPTR] as TInDevice).InputByte:=0;
    if HasTextExt(Filename) then rbPtrTextFile.Checked:=true else rbPtrBinFile.Checked:=true;
    Result:=true;
  except
    try FreeAndNil(PtrFile); except end;
    end;
  if not Result then ErrorDialog(_('Error opening file:')+sLineBreak+Filename);
  end;

function TMainForm.SelectPtrFile : boolean;
var
  s     : string;
begin
  s:=edPtr.Text; Result:=false;
  if OpenInputFile(_('PTR input from file'),_('Binary files')+'|*.bin|'+
      SetTextFileFilter(TextFiles)+_('all')+'|*.*','',s) then begin
    if OpenPtr(s) then begin
      AddToHistory(edPtr.Items,s);
      edPtr.ItemIndex:=0;
      Result:=true;
      end
    else edPtr.Text:='';
    end
  end;

procedure TMainForm.ReadFromPtr(Sender: TObject);
var
  AByte : byte;
  eof   : boolean;
begin
  if not assigned(PtrFile) then begin
    if not SelectPtrFile then begin
      btStopClick(Sender);
      (Sender as TInDevice).Reset;
      Exit;
      end;
    end;
  if assigned(PtrFile) then with PtrFile do begin
    eof:=Position>=Size;
    if not eof then Read(AByte,1)
    else if TrailerCount<TrailerLength then begin
      if rbPtrTextFile.Checked and (TrailerCount=0) then AByte:=$0C  // one form feed
      else AByte:=0;
      inc(TrailerCount);
      end;
    if TrailerCount<TrailerLength then (Sender as TInDevice).InputByte:=AByte;
    if (Position mod 10 =0) or eof then begin
      edPtrPos.Value:=Position;
      edPtrByte.Value:=AByte;
      end;
    end;
  end;

procedure TMainForm.edPtrCloseUp(Sender: TObject);
var
  ok : boolean;
  s  : string;
begin
  with edPtr do s:=Items[ItemIndex];
  if length(s)=0 then Exit;
  if assigned(PtrFile) then begin
    ok:=(PtrFile.TrailerCount>=TrailerLength)
      or ConfirmDialog(TopRightPos(btPtrOpen),Format(_('Close PTR input file:'+sLineBreak+'%s?'),[PtrFile.FName]));
    if ok then FreeAndNil(PtrFile);
    end;
  if not assigned(PtrFile) then begin
    if FileExists(s) then ok:=OpenPtr(s)
    else begin
      ErrorDialog(_('File not found:')+sLineBreak+s);
      ok:=false;
      end;
    with edPtr do if ok then Invalidate
    else ItemIndex:=-1;
    end
  else with edPtr do begin
    ItemIndex:=Items.IndexOf(LastPtr);
    end;
  end;

procedure TMainForm.edPtrDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
begin
  with edPtr do begin
    with Canvas do begin
      if (odSelected in State) and not (odComboBoxEdit in State) then Brush.Color:=clHighlight
      else Brush.Color:=clWindow;
      if (odComboBoxEdit in State) and assigned(PtrFile)  then Font.Color:=clWindowText;
      FillRect(Rect);
      TextOut(Rect.Left+2,Rect.Top,Items[Index]);
      end;
    end;
  end;

procedure TMainForm.edPtrDropDown(Sender: TObject);
begin
  LastPtr:=edPtr.Text;
  end;

procedure TMainForm.itmPtrOpenClick(Sender: TObject);
var
  ok : boolean;
begin
  if assigned(PtrFile) then begin
    ok:=(PtrFile.TrailerCount>=TrailerLength)
      or ConfirmDialog(TopRightPos(btPtrOpen),Format(_('Close PTR input file:'+sLineBreak+'%s?'),[PtrFile.FName]));
    if ok then FreeAndNil(PtrFile);
    end
  else ok:=true;
  if ok then ok:=SelectPtrFile;
  if not ok then edPtr.ItemIndex:=-1;
  end;

procedure TMainForm.itmPtrCloseClick(Sender: TObject);
begin
  if assigned(PtrFile) then PtrFile.TrailerCount:=TrailerLength; // set to end
//  FreeAndNil(PtrFile);
  edPtr.ItemIndex:=-1;
  end;

procedure TMainForm.itmPtrResetClick(Sender: TObject);
begin
  if assigned(PtrFile) then with PtrFile do begin
    Position:=0; TrailerCount:=0;
    edPtrPos.Value:=0; edPtrByte.Value:=0;
    (IoDevices[ioPTR] as TInDevice).InputByte:=0;
    end;
  end;

function TMainForm.OpenLpt (const Filename : string) : boolean;
begin
  Result:=false;
  try
    LptFile:=TFileStream.Create(Filename,fmCreate+fmShareDenyWrite);
    Result:=true;
  except
    try FreeAndNil(PtrFile); except end;
    end;
  end;

//-----------------------------------------------------------------------------
function TMainForm.SelectLptFile : boolean;
var
  s : string;
begin
  s:=edLpt.Text; Result:=false;
  if OpenOutputFile(_('LPT output to file'),_('all')+'|*.*','lst',s) then begin
    if OpenLpt(s) then edLpt.Text:=s
    else begin
      ErrorDialog(_('Error opening file:')+sLineBreak+s);
      edLpt.Text:=''; LptLine:='';
      mmLpt.Clear;
      Result:=true;
      end;
    end
  end;

procedure TMainForm.WriteToLPT (Sender: TObject; AByte : byte);
begin
  if not assigned(LptFile) then begin
    if SelectLptFile then begin
      btStopClick(Sender); Exit;
      end;
    end;
  if assigned(LptFile) then LptFile.Write(AByte,1);
  if AByte>=32 then LptLine:=LptLine+char(AByte)
  else if AByte=13 then begin   // CR
    mmLpt.Lines.Add(LptLine);
    LptLIne:='';
    end;
  end;

procedure TMainForm.btLptOpenClick(Sender: TObject);
begin
  if assigned(LptFile) and
      ConfirmDialog(TopRightPos(btLptOpen),Format(_('Close LPT output file:'+sLineBreak+'%s?'),[edLpt.Text])) then
    FreeAndNil(LptFile);
  if not assigned(LptFile) then SelectLptFile;
  end;

procedure TMainForm.btLptCloseClick(Sender: TObject);
begin
  FreeAndNil(LptFile);
  mmLpt.Clear;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.itWritePrgClick(Sender: TObject);
var
  First,Last,Mode : word;
  nf              : TNumMode;
  i               : integer;
  fLst            : TextFile;
begin
  with sgMemory.Selection do begin
    First:=Top; Last:=Bottom;
    end;
  Mode:=7; nf:=NumFormat;
  if WritePrgDialog.Execute(BottomRightPos(edPC),nf,First,Last,Mode) then with SaveDialog do begin
    Title:=_('Write program'); //'Programm schreiben';
    if length(ListName)>0 then InitialDir:=ExtractFilePath(ListName)
    else InitialDir:=WorkPath;
    DefaultExt:=LstExt;
    Filter:=_('Source listing')+'|*.'+LstExt+'|'+_('all')+'|*.*'; //Quelltext-Listing|*.'+LstExt+'|alle|*.*';
    FileName:='';
    if Execute then begin
      ListName:=FileName;
      AssignFile(fLst,ListName);
      Rewrite(fLst);
      for i:=First to Last do with Memory[i] do begin
        if Mode and 1 <>0 then Write(fLst,IntToVal(i,nf),'  ');
        if Mode and 2 <>0 then Write(fLst,IntToVal(Value,nf),'  ');
        if Mode and 4 <>0 then Write(fLst,Disassemble(i,Value,NumFormat,true));
        Writeln(fLst);
        end;
      CloseFile(fLst);
      end;
    end;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.ProgramLoad;
var
  i : integer;
begin
  for i:=0 to nPrgLoad-1 do Memory[i].Value:=PrgLoad[i];
  PC:=0;
  UpdateView;
  ShowCmd;
  RunProg;
  end;

procedure TMainForm.btPCClick(Sender: TObject);
begin
  sgMemory.Row:=PC;
  end;

procedure TMainForm.btPrgLoadClick(Sender: TObject);
begin
  if ValueDialog.Execute(BottomLeftPos(edSwitches),_('Front panel switches:'),NumFormat,FSwitches) then //'Frontplatten-Schalter:'
    ProgramLoad;
  end;

procedure TMainForm.itpSetPCClick(Sender: TObject);
begin
  PC:=sgMemory.Row;
  ShowCmd;
  sgMemory.Invalidate;
  end;

procedure TMainForm.btSwitchesToPCClick(Sender: TObject);
begin
  PC:=edSwitches.Value and $7FFF;
  ShowCmd;
  sgMemory.Invalidate;
  end;

procedure TMainForm.itpSeekClick(Sender: TObject);
var
  cmd : word;
begin
  with sgMemory do begin
    if ContentsCol then begin
      Row:=Memory[Row].Value;
      Invalidate;
      ContentsCol:=false;
      end
    else begin
      cmd:=Memory[Row].Value;
      // jmp, jsr, isz, dsz, lda, sta - abs, rel
      if (cmd and $E000<$6000) and (Hi(cmd) and 2=0) then begin
        if Hi(cmd) and 1=0 then Row:=Lo(cmd)
        else if Lo(cmd)>127 then Row:=Row+Lo(cmd)-256
          else Row:=Row+Lo(cmd);
        Invalidate;
        end;
      end;
    SaveMemPos(Row);
    end;
  end;

procedure TMainForm.itpGotoClick(Sender: TObject);
var
  adr : word;
begin
  with sgMemory do begin
    adr:=Row;
    if ValueDialog.Execute(CursorPos,_('Goto address:'),NumFormat,adr) then begin
      Row:=adr;
      SaveMemPos(Row);
      Invalidate;
      end;
    end;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.itpWatchClick(Sender: TObject);
begin
  AddWatch(sgMemory.Row);
  end;

procedure TMainForm.btAddWatchClick(Sender: TObject);
begin
  AddWatch(edAdr.Value);
  end;

procedure TMainForm.AddWatch (Addr : word);
begin
  AddToWatch(Addr,Memory[Addr].Value);
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.SetBreakpoint (Addr : word);
begin
  Include(Memory[Addr].BreakPoint,bpNormal); AddBreakpoint(Addr);
  if itmBreakpoints.Checked then BpWin.UpdateView(Breakpoints.Count);
  end;

procedure TMainForm.DelBreakpoint (Addr : word);
begin
  Exclude(Memory[Addr].BreakPoint,bpNormal); RemBreakpoint(Addr);
  if itmBreakpoints.Checked then BpWin.UpdateView(Breakpoints.Count);
  end;

function TMainForm.DeleteBreakpoint (const sAddr : string): integer;
var
  n : integer;
begin
  with Breakpoints do begin
    n:=IndexOf(sAddr);
    if n>=0 then begin
      Exclude(Memory[word(Objects[n])].BreakPoint,bpNormal);
      Delete(n);
      end;
    Result:=Count;
    end;
  end;

procedure TMainForm.itpToggleBpClick(Sender: TObject);
begin
  with Memory[sgMemory.Row] do if bpNormal in BreakPoint then
    DelBreakpoint(sgMemory.Row)
  else SetBreakpoint(sgMemory.Row);
  sgMemory.Invalidate;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.itmFrontpanelClick(Sender: TObject);
begin
  with itmFrontpanel do begin
    Checked:=not Checked;
    with NovaFrontpanel do begin
      Visible:=Checked;
      if Visible then begin
        Run:=FRun;
        Address:=PC;
        Carry:=FCarry<>0;
        Data:=edCMD.Value;
        Switches:=FSwitches;
        Application.ProcessMessages;
        end;
      end;
    end;
  end;

procedure TMainForm.btTraceClick(Sender: TObject);
begin
  with itmShowTrace do begin
    Checked:=not Checked;
    TraceWin.Visible:=Checked;
    end;
  end;

procedure TMainForm.btWatchClick(Sender: TObject);
begin
  with itmShowWatch do begin
    Checked:=not Checked;
    WatchWin.Visible:=Checked;
    if Checked then WatchWin.UpdateView(Watch.Count);
    end;
  end;

procedure TMainForm.itmBreakpointsClick(Sender: TObject);
begin
  with itmBreakpoints do begin
    Checked:=not Checked;
    BpWin.Visible:=Checked;
    if Checked then BpWin.UpdateView(Breakpoints.Count);
    end;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.itmGermanClick(Sender: TObject);
begin
  ChangeLanguage('de');
  itmGerman.Checked:=true;
  end;

procedure TMainForm.itmEnglishClick(Sender: TObject);
begin
  ChangeLanguage('en');
  itmEnglish.Checked:=true;
  end;

//-----------------------------------------------------------------------------
procedure TMainForm.btmemUpClick(Sender: TObject);
begin
  GotoMem(sgMemory.Row-edMemStep.Value);
  end;

procedure TMainForm.btmemDownClick(Sender: TObject);
begin
  GotoMem(sgMemory.Row+edMemStep.Value);
  end;

procedure TMainForm.edAC0KeyDown(Sender: TObject; var Key: Word;
 Shift: TShiftState);
begin
  if Key=VK_Return then edAc0Click(Sender);
  end;

procedure TMainForm.edAC1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_Return then edAc1Click(Sender);
  end;

procedure TMainForm.edAC2KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_Return then edAc2Click(Sender);
  end;

procedure TMainForm.edAC3KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_Return then edAc3Click(Sender);
  end;

procedure TMainForm.edAdrMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  r    : TRect;
  n    : integer;
  hi   : string;
begin
  if length(edAdr.Text)>0 then begin
    n:=edAdr.Value;
    hi:='=> '+IntToVal(Memory[n].Value,NumFormat)+' - '+DisAssemble(n,Memory[n].Value,NumFormat,true);
    with r do begin
      TopLeft:=edAdr.ClientToScreen(Point(20,edAdr.Height+2));
      Bottom:=Top+abs(Font.Height)+5;
      Right:=Left+MaxTextWidth(hi,Hintwin.Canvas)+10;
      end;
    HintWin.ActivateHint (r,hi);
    end;
  end;

procedure TMainForm.gbRegisterMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  HintWin.ReleaseHandle;
  end;

procedure TMainForm.edCarryKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_Return then edCarryClick(Sender);
  end;

end.
