(* Nova-Simulator
   ==============
   definitions and subroutines

    J. Rathlev, Schwentinental, (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   last modified:  August 2017
   *)

unit NovaUtils;

interface

uses System.SysUtils, NumberUtils;

type
  TDeviceName = record
    Name  : string;
    Value : word;
    end;

const
  // Sonderbefehle
  UIMUL = $76C1;
  UIDIV = $7641;
  INTEN = $607F;
  INTDS = $60BF;
  IORST = $65BF;
  PHALT = $663F;
  READS = $613F;
  INTA  = $633F;
  MSKO  = $643F;

  // I/O-Adressen
  ioMDV = $01;
  ioTTI = $08;
  ioTTO = $09;
  ioPTR = $0A;
  ioPTP = $0B;
  ioRTC = $0C;
  ioPLT = $0D;
  ioCDR = $0E;
  ioLPT = $0F;
  ioCPU = $3F;

  // I/O-Maskenbits
  mbTTO = 15;
  mbTTI = 14;
  mbPTP = 13;
  mbRTC = 13;
  mbPLT = 12;
  mbLPT = 12;
  mbPTR = 11;
  mbCDR = 10;

  // I/O tick counts
  tcTTO = 10;
  tcTTI = 5;
  tcPTP = 5;
  tcRTC = 0;
  tcPLT = 0;
  tcLPT = 5;
  tcPTR = 15;
  tcCDR = 5;

  NumWordDigits : array[TNumMode] of word = (5,4,6,16,5); // Stellenanzahl
  NumByteDigits : array[TNumMode] of word = (3,2,3,8,3); // Stellenanzahl

  AriLogInst : array [0..7] of string = ('COM','NEG','MOV','INC','ADC','SUB','ADD','AND');
  CarryInst  : array [0..3] of string = ('','Z','O','C');
  ShiftInst  : array [0..3] of string = ('','L','R','S');
  SkipInst   : array [0..7] of string = ('','SKP','SZC','SNC','SZR','SNR','SEZ','SBN');
  IoInst     : array [0..7] of string = ('NIO','DIA','DOA','DIB','DOB','DIC','DOC','SKP');
  CtrlInst   : array [0..3] of string = ('','S','C','P');
  IoSkpInst  : array [0..3] of string = ('BN','BZ','DN','DZ');
  JmpInst    : array [0..3] of string = ('JMP','JSR','ISZ','DSZ');
  MemInst    : array [0..1] of string = ('LDA','STA');
  SpecInst   : array [0..8] of string = ('MUL','DIV','INTEN','INTDS','IORST','HALT','READS','INTA','MSKO');

  DeviceCount = 7;
  DeviceNames : array [0..DeviceCount-1] of TDeviceName = (
    (Name : 'TTI'; Value : ioTTI),
    (Name : 'TTO'; Value : ioTTO),
    (Name : 'PTR'; Value : ioPTR),
    (Name : 'PTP'; Value : ioPTP),
    (Name : 'RTC'; Value : ioRTC),
    (Name : 'LPT'; Value : ioLPT),
    (Name : 'CPU'; Value : ioCPU)
    );

  // phys. Reihenfolge der Gerte
  NumberOfDevices = 6;
  DeviceChain : array [0..NumberOfDevices-1] of byte = (ioPTR,ioPTP,ioTTI,ioTTO,ioRTC,ioLpt);

function NumStrToVal (const S: string; ANumFormat : TNumMode) : word;
function IntToVal (Val : word; ANumFormat : TNumMode) : string; overload;
function IntToVal (Val,Dig : word; ANumFormat : TNumMode) : string; overload;
function IntToByte (Val : word; ANumFormat : TNumMode) : string;

function GetDeviceName (ADevice : word) : string;

function DisAssemble (Adr : integer; Value : word; ANumFormat : TNumMode; Align : boolean) : string;
function Assemble (Command : string; Addr : word; var Value : word) : boolean;

implementation

uses System.StrUtils, StringUtils;

{------------------------------------------------------------------}
function IntToVal (Val,Dig : word; ANumFormat : TNumMode) : string;
begin
  case ANumFormat of
  nmHex   : Result:=IntToHexStr(Val,Dig,0);
  nmOctaL : Result:=IntToOctal(Val,Dig,0);
  nmBin   : Result:=IntToBin(Val,Dig,0);
  nmZeroDec : Result:=IntToDecimal(Val,Dig,0,true);
  else Result:=IntToDecimal(Val,Dig,0);
    end;
  end;

function IntToVal (Val : word; ANumFormat : TNumMode) : string;
begin
  Result:=IntToVal(Val,NumWordDigits[ANumFormat],ANumFormat);
  end;

function IntToByte (Val : word; ANumFormat : TNumMode) : string;
begin
  Result:=IntToVal(Val,NumByteDigits[ANumFormat],ANumFormat);
  end;

{------------------------------------------------------------------}
function NumStrToVal (const S: string; ANumFormat : TNumMode) : word;
begin
  try
    case ANumFormat of
    nmHex   : Result:=HexStrToInt(S);
    nmOctaL : Result:=OctalStrToInt(S);
    nmBin   : Result:=BinStrToInt(S);
    else Result:=StrToInt(S);
      end;
  except
    Result:=0;
    end;
  end;

{------------------------------------------------------------------}
function GetDeviceName (ADevice : word) : string;
var
  i : integer;
begin
  for i:=0 to DeviceCount-1 do with DeviceNames[i] do if Value=ADevice then begin
    Result:=Name; Exit;
    end;
  Result:=IntToOctal(ADevice,0,0);
  end;

{------------------------------------------------------------------}
function DisAssemble (Adr : integer; Value : word; ANumFormat : TNumMode; Align : boolean) : string;
var
  s : string;
  ct,dv,n,i : word;
  k : integer;

  function MemRef(cmd : word) : string;
  var
    d : smallint;
    m : boolean;
  begin
    if cmd and $400 <>0 then Result:='@' else Result:='';
    n:=Hi(cmd) and 3;
    if n=0 then Result:=Result+IntToOctal(Lo(cmd),3,0)  // Page 0
    else begin
      d:=Lo(cmd);
      m:=false;
      if d>127 then begin
        d:=d-256; m:=true;
        end;
      if n=1 then begin                                 // rel. to PC
        if m then Result:=Result+'.-' else Result:=Result+'.+';
        Result:=Result+IntToOctal(abs(d),0,0);
        if Adr>=0 then Result:=Result+' ('+IntToVal(Adr+d,ANumFormat)+')';
        end
      else begin                                        // rel. to AC2/3
        if m then Result:=Result+'-';
        Result:=Result+IntToOctal(abs(d),0,0);
        if n=2 then Result:=Result+',2' else Result:=Result+',3';
        end;
      end;
    end;

begin
  if Align then k:=8 else k:=1;
  n:=(Hi(Value) and $18) shr 3;
  if (Value and $8000)<>0 then begin  // execute arithm inst.
    s:=AriLogInst[Hi(Value) and $7];
    s:=s+CarryInst[(Value and $30) shr 4];
    s:=s+ShiftInst[(Value and $C0) shr 6];
    if Value and $8 <>0 then s:=s+'#';
    s:=ExtSp(s+' ',k)+IntToStr((Hi(Value) and $60) shr 5)+','+IntToStr(n);
    n:=Value and 7;
    if n<>0 then s:=s+','+SkipInst[n];
    end
  else begin
    if Value and $6000 = 0 then begin // jmp, jsr, isz, dsz
      s:=ExtSp(JmpInst[n]+' ',k)+MemRef(Value);
      end
    else if Value and $6000 = $6000 then begin // In-Out
      i:=Hi(Value) and 7;
      ct:=(Value and $C0) shr 6;
      dv:=Value and $3F;
      s:='';
      if dv=$3F then begin // CPU instructions
        case Value of
        UIMUL : s:='MUL';
        UIDIV : s:='DIV';
        INTEN : s:='INTEN';
        INTDS : s:='INTDS';
        IORST : s:='IORST';
        PHALT : s:='HALT';
        else if ct=0 then begin
            case i of
            1 : s:=ExtSp('READS',k);
            3 : s:=ExtSp('INTA',k);
            4 : s:=ExtSp('MSKO',k);
              end;
            s:=s+IntToOctal(n,0,0);
            end;
          end;
        end;
      if length(s)=0 then begin
        s:=IoInst[i];
        if i=7 then s:=ExtSp(s+IoSkpInst[ct]+' ',k)+GetDeviceName(dv)
        else begin
          s:=ExtSp(s+CtrlInst[ct]+' ',k);
          if i=0 then s:=s+GetDeviceName(dv)
          else s:=s+IntToOctal(n,0,0)+','+GetDeviceName(dv);
          end
        end;
      end
    else begin  // lda, sta
      if Value and $6000 = $2000 then s:='LDA' else s:='STA';
      s:=ExtSp(s+' ',k)+IntToStr(n)+',';
      s:=s+MemRef(Value);
      end;
    end;
  Result:=s;
  end;

{------------------------------------------------------------------}
function Assemble (Command : string; Addr : word; var Value : word) : boolean;
var
  s : string;
  n : integer;
  w : word;

  function SeekMnemonic (Mnemonic : string; MList : array of string) : integer;
  var
    i : integer;
  begin
    Mnemonic:=UpperCase(Mnemonic);
    Result:=-1;
    for i:=Low(MList) to High(MList) do
      if Mnemonic=MList[i] then begin
        Result:=i; exit;
        end;
    end;

  function ReadAccu (var s : string) : integer;
  begin
    s:=TrimLeft(s);
    Result:=-1;
    if (length(s)>0) and (s[1]>='0') and (s[1]<='3') then Result:=ord(s[1])-48;
    ReadNxtStr(s,',');
    end;

  function CheckEndOfStr (var s : string) : boolean;
  begin
    delete(s,1,1);
    Result:=length(s)=0;
    end;

  function ReadDisp (s : string; var Value : word) : boolean;
  var
    minus : boolean;
  begin
    Result:=false;
    if length(s)=0 then exit;
    if s[1]='+' then delete(s,1,1);
    minus:=s[1]='-';
    if minus and CheckEndOfStr(s) then exit;
    try Value:=OctalStrToInt(s); except exit; end;
    if minus then begin
      if Value>128 then exit;
      Value:=$10000-Value;
      end
    else if n>127 then exit;
    Value:=Value and $FF;
    Result:=true;
    end;

  function ReadAddr (s : string; var Value : word) : boolean;
  var
    n : integer;
  begin
    Result:=false;
    s:=TrimLeft(s);
    if length(s)>0 then begin
      if s[1]='@' then begin  // indir.
        Value:=Value or $400;
        if CheckEndOfStr(s) then exit;
        end;
      n:=pos(',',s);
      if n>0 then begin   // rel. zu Ac2,Ac3
        if not ReadDisp(LeftStr(s,n-1),w) then exit;
        delete(s,1,n);
        Value:=Value or w;
        if length(s)=0 then exit;
        n:=ReadAccu(s);
        if n<2 then exit;
        Value:=Value or (n shl 8);
        Result:=true;
        end
      else if s[1]='.' then begin // rel. zu PC
        Value:=Value or $100;
        if CheckEndOfStr(s) then w:=0
        else if not ReadDisp(s,w) then exit;
        Value:=Value or w;
        Result:=true;
        end
      else begin   // direkt
        try w:=OctalStrToInt(s); except exit; end;
        if w>$FF then begin
          if (w<Addr-128) or (w>Addr+127) then exit;
          Value:=Value or $100 or ((w-Addr) and $FF);
          end
        else Value:=Value or (w and $FF);
        Result:=true;
        end;
      end;
    end;

  function ReadIoAdr (s : string; var Value : word) : boolean;
  var
    w : word;
    i : integer;
  begin
    Result:=false;
    s:=TrimLeft(s);
    i:=0;
    while (i<DeviceCount) and not AnsiSameText(s,DeviceNames[i].Name) do inc(i);
    if i<DeviceCount then Value:=Value or DeviceNames[i].Value
    else begin
      try w:=OctalStrToInt(s); except exit; end;
      Value:=Value or w;
      end;
    Result:=true;
    end;

begin
  n:=pos(';',Command);
  if n>0 then Command:=LeftStr(Command,n-1);
  Command:=Trim(Command);
  Command:=ReplChars(Command,Tab,Space);
  Value:=0;
  Result:=false;
  if length(Command)>=3 then begin
    s:=Command;
    n:=SeekMnemonic(ReadNxtStr(s,Space),SpecInst);
    if n>=0 then begin  // MUL,DIV,INTEN,...
      if n>=6 then begin
        case n of
        6 : Value:=READS;
        7 : Value:=INTA;
        8 : Value:=MSKO;
          end;
        n:=ReadAccu(s);
        if n<0 then exit;  // Fehler
        Value:=Value+n shl 11;
        end
      else case n of
      0 : Value:=UIMUL;
      1 : Value:=UIDIV;
      2 : Value:=INTEN;
      3 : Value:=INTDS;
      4 : Value:=IORST;
      5 : Value:=PHALT;
        end;
      end
    else begin
      s:=LeftStr(Command,3);   // die ersten drei Zeichen
      delete(Command,1,3);
      n:=SeekMnemonic(s,AriLogInst);
      if (n>=0) then begin  // COM,NEG,MOV,..
        Value:=$8000 or n shl 8;
        if (length(Command)>0) and (Command[1]<>Space) then begin
          n:=SeekMnemonic(copy(Command,1,1),CarryInst);
          if n>=0 then begin                 // Carry-Bed.
            Value:=Value or n shl 4;
            if CheckEndOfStr(Command) then exit;
            end;
          if Command[1]<>Space then begin
            n:=SeekMnemonic(copy(Command,1,1),ShiftInst);
            if n>=0 then begin                 // Shift
              Value:=Value or n shl 6;
              if CheckEndOfStr(Command) then exit;
              end;
            if Command[1]='#' then begin       // no load
              Value:=Value or $8;
              if CheckEndOfStr(Command) then exit;
              end
            else if Command[1]<>Space then exit;
            end;
          end;
        n:=ReadAccu(Command);                // AC-S
        if n<0 then exit;  // Fehler
        Value:=Value or n shl 13;
        n:=ReadAccu(Command);                // AC-D
        if n<0 then exit;  // Fehler
        Value:=Value or n shl 11;
        if length(Command)>0 then begin
          n:=SeekMnemonic(Command,SkipInst);
          if n>=0 then Value:=Value or n;
          end;
        end
      else begin
        n:=SeekMnemonic(s,MemInst);
        if n>=0 then begin  // LDA, STA
          if n=0 then Value:=$2000 else Value:=$4000;
          n:=ReadAccu(Command);
          if n<0 then exit;  // Fehler
          Value:=Value+n shl 11;
          if not ReadAddr(Command,Value) then exit;
          end
        else begin
          n:=SeekMnemonic(s,JmpInst);
          if n>=0 then begin  // JMP, JSR, ISZ, DSZ
            Value:=n shl 11;
            if not ReadAddr(Command,Value) then exit;
            end
          else begin
            n:=SeekMnemonic(s,IoInst);
            if n>=0 then begin  // NIO,DIA,DOA,...
              Value:=$6000 or n shl 8;
              w:=n;
              if w=7 then begin  // SKP
                n:=SeekMnemonic(copy(Command,1,2),IoSkpInst);
                if n<0 then exit;
                Value:=Value or n shl 6;
                Delete(Command,1,2);
                if not ReadIoAdr(Command,Value) then exit;
                end
              else begin
                if Command[1]<>Space then begin
                  n:=SeekMnemonic(copy(Command,1,1),CtrlInst);
                  if n<0 then exit;
                  Value:=Value or n shl 6;
                  if CheckEndOfStr(Command) then exit;
                  end;
                if w>0 then begin
                  n:=ReadAccu(Command);
                  if n<0 then exit;  // Fehler
                  Value:=Value+n shl 11;
                  end;
                if not ReadIoAdr(Command,Value) then exit;
                end;
              end
            end;
          end;
        end;
      end;
    end;
  Result:=true;
  end;

end.
